# (C) Copyright 2015,2016 Hewlett Packard Enterprise Development Company LP
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import json
import re
import urllib

import monasca_agent.collector.checks as checks


EVENT_TYPE = SOURCE_TYPE_NAME = 'rabbitmq'
QUEUE_TYPE = 'queues'
EXCHANGE_TYPE = 'exchanges'
NODE_TYPE = 'nodes'
MAX_DETAILED_QUEUES = 150
MAX_DETAILED_EXCHANGES = 100
MAX_DETAILED_NODES = 50

ALERT_THRESHOLD = 0.9
QUEUE_ATTRIBUTES = [
    # Path, Name
    ('active_consumers', 'active_consumers', float),
    ('consumers', 'consumers', float),
    ('memory', 'memory', float),

    ('messages', 'messages', float),
    ('messages_details/rate', 'messages.rate', float),

    ('messages_ready', 'messages.ready', float),
    ('messages_ready_details/rate', 'messages.ready_rate', float),

    ('messages_unacknowledged', 'messages.unacknowledged', float),
    ('messages_unacknowledged_details/rate', 'messages.unacknowledged_rate', float),

    ('message_stats/ack', 'messages.ack_count', float),
    ('message_stats/ack_details/rate', 'messages.ack_rate', float),

    ('message_stats/deliver', 'messages.deliver_count', float),
    ('message_stats/deliver_details/rate', 'messages.deliver_rate', float),

    ('message_stats/deliver_get', 'messages.deliver_get_count', float),
    ('message_stats/deliver_get_details/rate', 'messages.deliver_get_rate', float),

    ('message_stats/publish', 'messages.publish_count', float),
    ('message_stats/publish_details/rate', 'messages.publish_rate', float),

    ('message_stats/redeliver', 'messages.redeliver_count', float),
    ('message_stats/redeliver_details/rate', 'messages.redeliver_rate', float)]

EXCHANGE_ATTRIBUTES = [
    ('message_stats/publish_out', 'messages.published_count', float),
    ('message_stats/publish_out_details/rate', 'messages.published_rate', float),

    ('message_stats/publish_in', 'messages.received_count', float),
    ('message_stats/publish_in_details/rate', 'messages.received_rate', float)
]

NODE_ATTRIBUTES = [
    ('fd_used', 'fd_used', float),
    ('mem_used', 'mem_used', float),
    ('run_queue', 'run_queue', float),
    ('sockets_used', 'sockets_used', float),
    ('partitions', 'partitions', len)
]

ATTRIBUTES = {QUEUE_TYPE: QUEUE_ATTRIBUTES,
              EXCHANGE_TYPE: EXCHANGE_ATTRIBUTES,
              NODE_TYPE: NODE_ATTRIBUTES}

# whitelist of metrics to collect
DEFAULT_WHITELIST = {
    QUEUE_TYPE: ['messages',
                 'message_stats/deliver_details/rate',
                 'message_stats/publish_details/rate',
                 'message_stats/redeliver_details/rate'],
    EXCHANGE_TYPE: ['message_stats/publish_out',
                    'message_stats/publish_out_details/rate',
                    'message_stats/publish_in',
                    'message_stats/publish_in_details/rate'],
    NODE_TYPE: ['fd_used',
                'mem_used',
                'run_queue',
                'sockets_used']
}

DIMENSIONS_MAP = {
    'queues': {'name': 'queue',
               'vhost': 'vhost',
               'policy': 'policy'},
    'exchanges': {'name': 'exchange',
                  'vhost': 'vhost',
                  'type': 'type'},
    'nodes': {'name': 'node'}
}

METRIC_SUFFIX = {QUEUE_TYPE: "queue", EXCHANGE_TYPE: "exchange", NODE_TYPE: "node"}


class RabbitMQ(checks.AgentCheck):

    """This check is for gathering statistics from the RabbitMQ

    Management Plugin (http://www.rabbitmq.com/management.html)
    """

    def __init__(self, name, init_config, agent_config, instances=None):
        checks.AgentCheck.__init__(self, name, init_config, agent_config, instances)
        self.already_alerted = []

    @staticmethod
    def _get_config(instance):
        # make sure 'rabbitmq_api_url; is present
        if 'rabbitmq_api_url' not in instance:
            raise Exception('Missing "rabbitmq_api_url" in RabbitMQ config.')

        # get parameters
        base_url = instance['rabbitmq_api_url']
        if not base_url.endswith('/'):
            base_url += '/'
        username = instance.get('rabbitmq_user', 'guest')
        password = instance.get('rabbitmq_pass', 'guest')

        # Limit of queues/nodes to collect metrics from
        max_detailed = {
            QUEUE_TYPE: int(instance.get('max_detailed_queues', MAX_DETAILED_QUEUES)),
            EXCHANGE_TYPE: int(instance.get('max_detailed_exchanges', MAX_DETAILED_EXCHANGES)),
            NODE_TYPE: int(instance.get('max_detailed_nodes', MAX_DETAILED_NODES)),
        }

        # List of queues/nodes to collect metrics from
        specified = {
            QUEUE_TYPE: {
                'explicit': instance.get('queues', []),
                'regexes': instance.get('queues_regexes', []),
            },
            EXCHANGE_TYPE: {
                'explicit': instance.get('exchanges', []),
                'regexes': instance.get('exchanges_regexes', []),
            },
            NODE_TYPE: {
                'explicit': instance.get('nodes', []),
                'regexes': instance.get('nodes_regexes', []),
            },
        }

        for object_type, filters in specified.items():
            for filter_type, filter_objects in filters.items():
                if type(filter_objects) != list:
                    raise TypeError(
                        "{0} / {0}_regexes parameter must be a list".format(object_type))

        # setup urllib2 for Basic Auth
        auth_handler = urllib.request.HTTPBasicAuthHandler()
        auth_handler.add_password(
            realm='RabbitMQ Management', uri=base_url, user=username, passwd=password)
        opener = urllib.request.build_opener(auth_handler)
        urllib.request.install_opener(opener)

        return base_url, max_detailed, specified

    def check(self, instance):
        base_url, max_detailed, specified = self._get_config(instance)

        self.get_stats(instance,
                       base_url,
                       QUEUE_TYPE,
                       max_detailed[QUEUE_TYPE],
                       specified[QUEUE_TYPE])
        self.get_stats(instance,
                       base_url,
                       NODE_TYPE,
                       max_detailed[NODE_TYPE],
                       specified[NODE_TYPE])
        self.get_stats(instance,
                       base_url,
                       EXCHANGE_TYPE,
                       max_detailed[EXCHANGE_TYPE],
                       specified[EXCHANGE_TYPE])

    @staticmethod
    def _get_data(url):
        try:
            data = json.loads(urllib.request.urlopen(url).read())
        except urllib.error.URLError as e:
            raise Exception('Cannot open RabbitMQ API url: %s %s' % (url, str(e)))
        except ValueError as e:
            raise Exception('Cannot parse JSON response from API url: %s %s' % (url, str(e)))
        return data

    def get_stats(self, instance, base_url, object_type, max_detailed, filters):
        """instance: the check instance

        base_url: the url of the rabbitmq management api (e.g. http://localhost:15672/api)
        object_type: either QUEUE_TYPE, EXCHANGE_TYPE or NODE_TYPE
        max_detailed: the limit of objects to collect for this type
        filters: explicit or regexes filters of specified queues or nodes
                (specified in the yaml file)
        """
        data = self._get_data(urllib.parse.urljoin(base_url, object_type))
        # Make a copy of this list as we will remove items from it at each iteration
        explicit_filters = list(filters['explicit'])
        regex_filters = filters['regexes']

        """data is a list of nodes or queues:

        data = [
            { 'status': 'running', 'node': 'rabbit@host', 'name': 'queue1', 'consumers': 0,
            'vhost': '/', 'backing_queue_status': {'q1': 0, 'q3': 0, 'q2': 0, 'q4': 0,
            'avg_ack_egress_rate': 0.0, 'ram_msg_count': 0, 'ram_ack_count': 0, 'len': 0,
            'persistent_count': 0, 'target_ram_count': 'infinity', 'next_seq_id': 0,
            'delta': ['delta', 'undefined', 0, 'undefined'], 'pending_acks': 0,
            'avg_ack_ingress_rate': 0.0, 'avg_egress_rate': 0.0, 'avg_ingress_rate': 0.0},
            'durable': True, 'idle_since': '2013-10-03 13:38:18', 'exclusive_consumer_tag': '',
            'arguments': {}, 'memory': 10956, 'policy': '', 'auto_delete': False},
            {'status': 'running', 'node': 'rabbit@host, 'name': 'queue10', 'consumers': 0,
            'vhost': '/', 'backing_queue_status': {'q1': 0, 'q3': 0, 'q2': 0, 'q4': 0,
            'avg_ack_egress_rate': 0.0, 'ram_msg_count': 0, 'ram_ack_count': 0, 'len': 0,
            'persistent_count': 0, 'target_ram_count': 'infinity', 'next_seq_id': 0,
            'delta': ['delta', 'undefined', 0, 'undefined'], 'pending_acks': 0,
            'avg_ack_ingress_rate': 0.0, 'avg_egress_rate': 0.0, 'avg_ingress_rate': 0.0},
            'durable': True, 'idle_since': '2013-10-03 13:38:18', 'exclusive_consumer_tag': '',
            'arguments': {}, 'memory': 10956, 'policy': '', 'auto_delete': False},
            {'status': 'running', 'node': 'rabbit@host', 'name': 'queue11', 'consumers': 0,
            'vhost': '/', 'backing_queue_status': {'q1': 0, 'q3': 0, 'q2': 0, 'q4': 0,
            'avg_ack_egress_rate': 0.0, 'ram_msg_count': 0, 'ram_ack_count': 0, 'len': 0,
            'persistent_count': 0, 'target_ram_count': 'infinity', 'next_seq_id': 0,
            'delta': ['delta', 'undefined', 0, 'undefined'], 'pending_acks': 0,
            'avg_ack_ingress_rate': 0.0, 'avg_egress_rate': 0.0, 'avg_ingress_rate': 0.0},
            'durable': True, 'idle_since': '2013-10-03 13:38:18', 'exclusive_consumer_tag': '',
            'arguments': {}, 'memory': 10956, 'policy': '', 'auto_delete': False},
            ...
        ]
        """
        if len(explicit_filters) > max_detailed:
            raise Exception("The maximum number of %s you can specify is %d." %
                            (object_type, max_detailed))

        # a list of queues/nodes is specified. We process only those
        if explicit_filters or regex_filters:
            matching_lines = []
            for data_line in data:
                name = data_line.get("name")
                if self._is_matching_stat(name, explicit_filters, regex_filters):
                    matching_lines.append(data_line)
                    continue

                # Absolute names work only for queues
                if object_type != QUEUE_TYPE:
                    continue
                absolute_name = '%s/%s' % (data_line.get("vhost"), name)
                if self._is_matching_stat(absolute_name, explicit_filters, regex_filters):
                    matching_lines.append(data_line)
                    continue

            data = matching_lines

        if len(data) > max_detailed:
            self.log.warning(
                "Too many %s to fetch.  Increase max_detailed_ in the config "
                "or results will be truncated." %
                object_type)

        for data_line in data[:max_detailed]:
            # We truncate the list of nodes/queues if it's above the limit
            self._get_metrics(data_line, object_type, instance)

    def _is_matching_stat(self, name, explicit_filters, regex_filters):
        if name in explicit_filters:
            return True
        for p in regex_filters:
            match = re.search(p, name)
            if match:
                return True
        return False

    def _get_metrics(self, data, object_type, instance):
        whitelist = instance.get('whitelist', {})
        object_whitelist = whitelist.get(object_type, DEFAULT_WHITELIST[object_type])

        dimensions_list = DIMENSIONS_MAP[object_type].copy()
        dimensions = self._set_dimensions({'component': 'rabbitmq', 'service': 'rabbitmq'},
                                          instance)
        for d in dimensions_list:
            dim = data.get(d, None)
            if dim not in [None, ""]:
                dimensions[dimensions_list[d]] = dim

        for attribute, metric_name, operation in ATTRIBUTES[object_type]:
            # Check if the metric should be collected
            if attribute not in object_whitelist:
                continue

            # Walk down through the data path, e.g. foo/bar => d['foo']['bar']
            root = data
            keys = attribute.split('/')
            for path in keys[:-1]:
                root = root.get(path, {})

            value = root.get(keys[-1], None)
            if value is None:
                value = 0.0
            try:
                self.log.debug(
                    "Collected data for %s: metric name: %s: value: %f dimensions: %s" %
                    (object_type, metric_name, operation(value), str(dimensions)))
                self.gauge(
                    'rabbitmq.%s.%s' %
                    (METRIC_SUFFIX[object_type],
                     metric_name),
                    operation(value),
                    dimensions=dimensions)
            except ValueError:
                self.log.exception("Caught ValueError for %s %s = %s  with dimensions: %s" % (
                    METRIC_SUFFIX[object_type], attribute, value, dimensions))



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import random
import time
import unittest

import nose.tools as nt

from monasca_agent.common.aggregator import MetricsAggregator


class TestUnitMonascaStatsd(unittest.TestCase):

    @staticmethod
    def sort_metrics(metrics):
        def sort_by(m):
            return (m['metric'], ','.join(m['dimensions'] or []))

        return sorted(metrics, key=sort_by)

    @staticmethod
    def sort_events(metrics):
        def sort_by(m):
            return (m['title'], m['text'], ','.join(m.get('tags', None) or []))

        return sorted(metrics, key=sort_by)

    def test_counter_normalization(self):
        stats = MetricsAggregator('myhost', interval=10)

        # Assert counters are normalized.
        stats.submit_packets('int:1|c')
        stats.submit_packets('int:4|c')
        stats.submit_packets('int:15|c')

        stats.submit_packets('float:5|c')

        metrics = self.sort_metrics(stats.flush())
        assert len(metrics) == 2

        floatc, intc = metrics

        nt.assert_equal(floatc['metric'], 'float')
        nt.assert_equal(floatc['points'][0][1], 0.5)
        nt.assert_equal(floatc['host'], 'myhost')

        nt.assert_equal(intc['metric'], 'int')
        nt.assert_equal(intc['points'][0][1], 2)
        nt.assert_equal(intc['host'], 'myhost')

    def test_histogram_normalization(self):
        stats = MetricsAggregator('myhost', interval=10)
        for i in range(5):
            stats.submit_packets('h1:1|h')
        for i in range(20):
            stats.submit_packets('h2:1|h')

        metrics = self.sort_metrics(stats.flush())
        _, _, h1count, _, _, _, _, h2count, _, _ = metrics

        nt.assert_equal(h1count['points'][0][1], 0.5)
        nt.assert_equal(h2count['points'][0][1], 2)

    def test_tags(self):
        stats = MetricsAggregator('myhost')
        stats.submit_packets('gauge:1|c')
        stats.submit_packets('gauge:2|c|@1')
        stats.submit_packets('gauge:4|c|#tag1,tag2')
        stats.submit_packets('gauge:8|c|#tag2,tag1')  # Should be the same as above
        stats.submit_packets('gauge:16|c|#tag3,tag4')

        metrics = self.sort_metrics(stats.flush())

        assert len(metrics) == 3
        first, second, third = metrics

        nt.assert_equal(first['metric'], 'gauge')
        nt.assert_equal(first['dimensions'], None)
        nt.assert_equal(first['points'][0][1], 3)
        nt.assert_equal(first['host'], 'myhost')

        nt.assert_equal(second['metric'], 'gauge')
        nt.assert_equal(second['dimensions'], ('tag1', 'tag2'))
        nt.assert_equal(second['points'][0][1], 12)
        nt.assert_equal(second['host'], 'myhost')

        nt.assert_equal(third['metric'], 'gauge')
        nt.assert_equal(third['dimensions'], ('tag3', 'tag4'))
        nt.assert_equal(third['points'][0][1], 16)
        nt.assert_equal(third['host'], 'myhost')

    def test_counter(self):
        stats = MetricsAggregator('myhost')

        # Track some counters.
        stats.submit_packets('my.first.counter:1|c')
        stats.submit_packets('my.first.counter:5|c')
        stats.submit_packets('my.second.counter:1|c')
        stats.submit_packets('my.third.counter:3|c')

        # Ensure they roll up nicely.
        metrics = self.sort_metrics(stats.flush())
        assert len(metrics) == 3

        first, second, third = metrics
        nt.assert_equals(first['metric'], 'my.first.counter')
        nt.assert_equals(first['points'][0][1], 6)
        nt.assert_equals(first['host'], 'myhost')

        nt.assert_equals(second['metric'], 'my.second.counter')
        nt.assert_equals(second['points'][0][1], 1)

        nt.assert_equals(third['metric'], 'my.third.counter')
        nt.assert_equals(third['points'][0][1], 3)

        # Ensure that counters reset to zero.
        metrics = self.sort_metrics(stats.flush())
        first, second, third = metrics
        nt.assert_equals(first['metric'], 'my.first.counter')
        nt.assert_equals(first['points'][0][1], 0)
        nt.assert_equals(second['metric'], 'my.second.counter')
        nt.assert_equals(second['points'][0][1], 0)
        nt.assert_equals(third['metric'], 'my.third.counter')
        nt.assert_equals(third['points'][0][1], 0)

    def test_sampled_counter(self):

        # Submit a sampled counter.
        stats = MetricsAggregator('myhost')
        stats.submit_packets('sampled.counter:1|c|@0.5')
        metrics = stats.flush()
        assert len(metrics) == 1
        m = metrics[0]
        assert m['metric'] == 'sampled.counter'
        nt.assert_equal(m['points'][0][1], 2)

    def test_gauge(self):
        stats = MetricsAggregator('myhost')

        # Track some counters.
        stats.submit_packets('my.first.gauge:1|g')
        stats.submit_packets('my.first.gauge:5|g')
        stats.submit_packets('my.second.gauge:1.5|g')

        # Ensure that gauges roll up correctly.
        metrics = self.sort_metrics(stats.flush())
        assert len(metrics) == 2

        first, second = metrics

        nt.assert_equals(first['metric'], 'my.first.gauge')
        nt.assert_equals(first['points'][0][1], 5)
        nt.assert_equals(first['host'], 'myhost')

        nt.assert_equals(second['metric'], 'my.second.gauge')
        nt.assert_equals(second['points'][0][1], 1.5)

        # Ensure that old gauges get dropped due to old timestamps
        stats.gauge('my.first.gauge', 5)
        stats.gauge('my.first.gauge', 1, timestamp=1000000000)
        stats.gauge('my.second.gauge', 20, timestamp=1000000000)

        metrics = self.sort_metrics(stats.flush())
        assert len(metrics) == 1

        first = metrics[0]

        nt.assert_equals(first['metric'], 'my.first.gauge')
        nt.assert_equals(first['points'][0][1], 5)
        nt.assert_equals(first['host'], 'myhost')

    def test_sets(self):
        stats = MetricsAggregator('myhost')
        stats.submit_packets('my.set:10|s')
        stats.submit_packets('my.set:20|s')
        stats.submit_packets('my.set:20|s')
        stats.submit_packets('my.set:30|s')
        stats.submit_packets('my.set:30|s')
        stats.submit_packets('my.set:30|s')

        # Assert that it's treated normally.
        metrics = stats.flush()
        nt.assert_equal(len(metrics), 1)
        m = metrics[0]
        nt.assert_equal(m['metric'], 'my.set')
        nt.assert_equal(m['points'][0][1], 3)

        # Assert there are no more sets
        assert not stats.flush()

    def test_string_sets(self):
        stats = MetricsAggregator('myhost')
        stats.submit_packets('my.set:string|s')
        stats.submit_packets('my.set:sets|s')
        stats.submit_packets('my.set:sets|s')
        stats.submit_packets('my.set:test|s')
        stats.submit_packets('my.set:test|s')
        stats.submit_packets('my.set:test|s')

        # Assert that it's treated normally.
        metrics = stats.flush()
        nt.assert_equal(len(metrics), 1)
        m = metrics[0]
        nt.assert_equal(m['metric'], 'my.set')
        nt.assert_equal(m['points'][0][1], 3)

        # Assert there are no more sets
        assert not stats.flush()

    def test_rate(self):
        stats = MetricsAggregator('myhost')
        stats.submit_packets('my.rate:10|_dd-r')
        # Sleep 1 second so the time interval > 0
        time.sleep(1)
        stats.submit_packets('my.rate:40|_dd-r')

        # Check that the rate is calculated correctly
        metrics = stats.flush()
        nt.assert_equal(len(metrics), 1)
        m = metrics[0]
        nt.assert_equals(m['metric'], 'my.rate')
        nt.assert_equals(m['points'][0][1], 30)

        # Assert that no more rates are given
        assert not stats.flush()

    def test_gauge_sample_rate(self):
        stats = MetricsAggregator('myhost')

        # Submit a sampled gauge metric.
        stats.submit_packets('sampled.gauge:10|g|@0.1')

        # Assert that it's treated normally.
        metrics = stats.flush()
        nt.assert_equal(len(metrics), 1)
        m = metrics[0]
        nt.assert_equal(m['metric'], 'sampled.gauge')
        nt.assert_equal(m['points'][0][1], 10)

    def test_histogram(self):
        stats = MetricsAggregator('myhost')

        # Sample all numbers between 1-100 many times. This
        # means our percentiles should be relatively close to themselves.
        percentiles = range(100)
        random.shuffle(percentiles)  # in place
        for i in percentiles:
            for j in range(20):
                for type_ in ['h', 'ms']:
                    m = 'my.p:%s|%s' % (i, type_)
                    stats.submit_packets(m)

        metrics = self.sort_metrics(stats.flush())

        def assert_almost_equal(i, j, e=1):
            # Floating point math?
            assert abs(i - j) <= e, "%s %s %s" % (i, j, e)

        nt.assert_equal(len(metrics), 5)
        p95, pavg, pcount, pmax, pmed = self.sort_metrics(metrics)
        nt.assert_equal(p95['metric'], 'my.p.95percentile')
        assert_almost_equal(p95['points'][0][1], 95, 10)
        assert_almost_equal(pmax['points'][0][1], 99, 1)
        assert_almost_equal(pmed['points'][0][1], 50, 2)
        assert_almost_equal(pavg['points'][0][1], 50, 2)
        assert_almost_equal(pcount['points'][0][1], 4000, 0)  # 100 * 20 * 2
        nt.assert_equals(p95['host'], 'myhost')

        # Ensure that histograms are reset.
        metrics = self.sort_metrics(stats.flush())
        assert not metrics

    def test_sampled_histogram(self):
        # Submit a sampled histogram.
        stats = MetricsAggregator('myhost')
        stats.submit_packets('sampled.hist:5|h|@0.5')

        # Assert we scale up properly.
        metrics = self.sort_metrics(stats.flush())
        p95, pavg, pcount, pmax, pmed = self.sort_metrics(metrics)

        nt.assert_equal(pcount['points'][0][1], 2)
        for p in [p95, pavg, pmed, pmax]:
            nt.assert_equal(p['points'][0][1], 5)

    def test_batch_submission(self):
        # Submit a sampled histogram.
        stats = MetricsAggregator('myhost')
        metrics = [
            'counter:1|c',
            'counter:1|c',
            'gauge:1|g'
        ]
        packet = "\n".join(metrics)
        stats.submit_packets(packet)

        metrics = self.sort_metrics(stats.flush())
        nt.assert_equal(2, len(metrics))
        counter, gauge = metrics
        assert counter['points'][0][1] == 2
        assert gauge['points'][0][1] == 1

    def test_bad_packets_throw_errors(self):
        packets = [
            'missing.value.and.type',
            'missing.type:2',
            'missing.value|c',
            '2|c',
            'unknown.type:2|z',
            'string.value:abc|c',
            'string.sample.rate:0|c|@abc',
            # Bad event-like packets
            '_ev{1,2}:bad_header'
            '_e{1,}:invalid|headers',
            '_e:missing|size|headers',
            '_e:{1,1}:t|t|t:bad_meta|h',
        ]

        stats = MetricsAggregator('myhost')
        for packet in packets:
            try:
                stats.submit_packets(packet)
            except Exception:
                assert True
            else:
                assert False, 'invalid : %s' % packet

    def test_metrics_expiry(self):
        # Ensure metrics eventually expire and stop submitting.
        stats = MetricsAggregator('myhost', expiry_seconds=1)
        stats.submit_packets('test.counter:123|c')

        # Ensure points keep submitting
        assert stats.flush()
        assert stats.flush()
        time.sleep(0.5)
        assert stats.flush()

        # Now sleep for longer than the expiry window and ensure
        # no points are submitted
        time.sleep(2)
        m = stats.flush()
        assert not m, str(m)

        # If we submit again, we're all good.
        stats.submit_packets('test.counter:123|c')
        assert stats.flush()

    def test_histogram_counter(self):
        # Test whether histogram.count == increment
        # same deal with a sample rate
        cnt = 100000
        for run in [1, 2]:
            stats = MetricsAggregator('myhost')
            for i in range(cnt):
                if run == 2:
                    stats.submit_packets('test.counter:1|c|@0.5')
                    stats.submit_packets('test.hist:1|ms|@0.5')
                else:
                    stats.submit_packets('test.counter:1|c')
                    stats.submit_packets('test.hist:1|ms')
            metrics = self.sort_metrics(stats.flush())
            assert len(metrics) > 0

            nt.assert_equal([m['points'][0][1]
                             for m in metrics if m['metric'] == 'test.counter'], [cnt * run])
            nt.assert_equal([m['points'][0][1]
                             for m in metrics if m['metric'] == 'test.hist.count'], [cnt * run])

    def test_scientific_notation(self):
        stats = MetricsAggregator('myhost', interval=10)

        stats.submit_packets('test.scinot:9.512901e-05|g')
        metrics = self.sort_metrics(stats.flush())

        assert len(metrics) == 1
        ts, val = metrics[0].get('points')[0]
        nt.assert_almost_equal(val, 9.512901e-05)

    def test_event_tags(self):
        stats = MetricsAggregator('myhost')
        stats.submit_packets('_e{6,4}:title1|text')
        stats.submit_packets('_e{6,4}:title2|text|#t1')
        stats.submit_packets('_e{6,4}:title3|text|#t1,t2:v2,t3,t4')
        stats.submit_packets('_e{6,4}:title4|text|k:key|p:normal|#t1,t2')

        events = self.sort_events(stats.flush_events())

        assert len(events) == 4
        first, second, third, fourth = events

        try:
            first['dimensions']
        except Exception:
            assert True
        else:
            assert False, "event['tags'] shouldn't be defined when no tags aren't explicited in the packet"
        nt.assert_equal(first['title'], 'title1')
        nt.assert_equal(first['text'], 'text')

        nt.assert_equal(second['title'], 'title2')
        nt.assert_equal(second['text'], 'text')
        nt.assert_equal(second['dimensions'], sorted(['t1']))

        nt.assert_equal(third['title'], 'title3')
        nt.assert_equal(third['text'], 'text')
        nt.assert_equal(third['dimensions'], sorted(['t1', 't2:v2', 't3', 't4']))

        nt.assert_equal(fourth['title'], 'title4')
        nt.assert_equal(fourth['text'], 'text')
        nt.assert_equal(fourth['aggregation_key'], 'key')
        nt.assert_equal(fourth['priority'], 'normal')
        nt.assert_equal(fourth['dimensions'], sorted(['t1', 't2']))

    def test_event_title(self):
        stats = MetricsAggregator('myhost')
        stats.submit_packets('_e{0,4}:|text')
        stats.submit_packets(u'_e{9,4}:2intitul��|text')
        stats.submit_packets('_e{14,4}:3title content|text')
        stats.submit_packets('_e{14,4}:4title|content|text')
        stats.submit_packets('_e{13,4}:5title\\ntitle|text')  # \n stays escaped

        events = self.sort_events(stats.flush_events())

        assert len(events) == 5
        first, second, third, fourth, fifth = events

        nt.assert_equal(first['title'], '')
        nt.assert_equal(second['title'], u'2intitul��')
        nt.assert_equal(third['title'], '3title content')
        nt.assert_equal(fourth['title'], '4title|content')
        nt.assert_equal(fifth['title'], '5title\\ntitle')

    def test_event_text(self):
        stats = MetricsAggregator('myhost')
        stats.submit_packets('_e{2,0}:t1|')
        stats.submit_packets('_e{2,12}:t2|text|content')
        stats.submit_packets('_e{2,23}:t3|First line\\nSecond line')  # \n is a newline
        stats.submit_packets(u'_e{2,19}:t4|��� �����U �����U ����u T0�� ���')  # utf-8 compliant

        events = self.sort_events(stats.flush_events())

        assert len(events) == 4
        first, second, third, fourth = events

        nt.assert_equal(first['text'], '')
        nt.assert_equal(second['text'], 'text|content')
        nt.assert_equal(third['text'], 'First line\nSecond line')
        nt.assert_equal(fourth['text'], u'��� �����U �����U ����u T0�� ���')


if __name__ == "__main__":
    unittest.main()



# Copyright 2016 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging

import requests

from monasca_agent.collector import checks
from monasca_agent.common import util

from monasca_setup.detection.plugins import kibana as kibana_setup

LOG = logging.getLogger(__name__)

_ONE_MB = (1024 * 1024) * 1.0
_LOAD_TIME_SERIES = ['1m', '5m', '15m']


class Kibana(checks.AgentCheck):
    def get_library_versions(self):
        try:
            import yaml
            version = yaml.__version__
        except ImportError:
            version = "Not Found"
        except AttributeError:
            version = "Unknown"

        return {"PyYAML": version}

    def check(self, instance):
        config_url = self.init_config.get('url', None)

        if config_url is None:
            raise Exception('An url to kibana must be specified')

        instance_metrics = instance.get('metrics', None)
        if not instance_metrics:
            LOG.warning('All metrics have been disabled in configuration '
                        'file, nothing to do.')
            return

        version = self._get_kibana_version(config_url)
        dimensions = self._set_dimensions({'version': version}, instance)

        LOG.debug('Kibana version %s', version)

        try:
            stats = self._get_data(config_url)
        except Exception as ex:
            LOG.error('Error while trying to get stats from Kibana[%s]' %
                      config_url)
            LOG.exception(ex)
            return

        if not stats:
            LOG.warning('No stats data was collected from kibana')
            return

        self._process_metrics(stats, dimensions, instance_metrics)

    def _get_data(self, url):
        return requests.get(
            url=url,
            headers=util.headers(self.agent_config)
        ).json()

    def _process_metrics(self, stats, dimensions, instance_metrics):
        # collect from instance which metrics should be checked
        actual_metrics = {kibana_setup.get_metric_name(k): v for k, v in
                          stats.get('metrics', {}).items()}
        instance_url = self.init_config.get('url')

        for metric in actual_metrics.keys():
            if metric not in instance_metrics:
                LOG.debug('%s has been disabled for %s check' % (
                    metric, instance_url))
                continue
            else:
                self._process_metric(metric,
                                     actual_metrics.get(metric),
                                     dimensions)

    def _process_metric(self, metric, stats, dimensions):
        LOG.debug('Processing metric %s' % metric)

        metric_name = self.normalize(metric, 'kibana')

        if metric in ['heap_size', 'heap_used']:
            metric_name = '%s_mb' % metric_name
        elif metric in ['resp_time_max', 'resp_time_avg']:
            metric_name = '%s_ms' % metric_name

        for item in stats:
            timestamp = int(item[0]) / 1000.0
            measurements = item[1]
            cleaned_metric_name = metric_name

            if not isinstance(measurements, list):
                # only load comes as list in measurements
                measurements = [measurements]

            for it, measurement in enumerate(measurements):
                if measurement is None:
                    LOG.debug('Measurement for metric %s at %d was not '
                              'returned from kibana server, skipping'
                              % (metric_name, timestamp))
                    continue

                if metric in ['heap_size', 'heap_used']:
                    measurement /= _ONE_MB
                elif metric == 'load':
                    load_sub_metric = _LOAD_TIME_SERIES[it]
                    cleaned_metric_name = '%s_avg_%s' % (metric_name,
                                                         load_sub_metric)

                LOG.debug('Reporting %s as gauge with value %f'
                          % (cleaned_metric_name, measurement))

                self.gauge(
                    metric=cleaned_metric_name,
                    value=measurement,
                    dimensions=dimensions,
                    timestamp=timestamp
                )

    def _get_kibana_version(self, url):
        return requests.head(url=url).headers['kbn-version']



# (C) Copyright 2017 Hewlett Packard Enterprise Development LP
# coding=utf-8

from monasca_agent.collector.checks import utils

kubernetes_connector = utils.KubernetesConnector(3)
print(kubernetes_connector.get_agent_pod_host(return_host_name=True))



# (C) Copyright 2015-2016 Hewlett Packard Enterprise Development Company LP
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from monasca_agent.collector.checks import AgentCheck
from os.path import abspath
from os.path import exists
from os.path import join

from os import access
from os import stat
from os import walk

from os import X_OK

import logging

log = logging.getLogger(__name__)
missing_dir_instance = set()
non_exist_dir = set()
no_access_sub_dir = set()
no_stat_file = set()


class DirectoryCheck(AgentCheck):

    """This check is for monitoring and reporting metrics on the provided directory

    WARNING: the user/group that mon-agent runs as must have access to stat
    the files in the desired directory

    Config options:
        "directory" - string, the directory to gather stats for. required
    """

    def check(self, instance):
        global missing_dir_instance
        global non_exist_dir
        if "directory" not in instance:
            if instance['dimensions']['service'] not in missing_dir_instance:
                missing_dir_instance.add(instance['dimensions']['service'])
                log.error('DirectoryCheck: missing "directory" in config')
        else:
            directory = instance["directory"]
            abs_directory = abspath(directory)
            if not exists(abs_directory) and abs_directory not in non_exist_dir:
                non_exist_dir.add(abs_directory)
                log.error("DirectoryCheck: the directory (%s) does not exist" % abs_directory)

            dimensions = self._set_dimensions({"path": directory}, instance)
            self._get_stats(abs_directory, dimensions)

    def _get_stats(self, directory_name, dimensions):
        global no_access_sub_dir
        global no_stat_file
        directory_bytes = 0
        directory_files = 0
        for root, dirs, files in walk(directory_name):
            for directory in dirs:
                directory_root = join(root, directory)
                if not access(directory_root, X_OK) and directory_root not in no_access_sub_dir:
                    no_access_sub_dir.add(directory_root)
                    log.warn("DirectoryCheck: could not access sub directory "
                             "{}".format(directory_root))
            for filename in files:
                filename = join(root, filename)
                try:
                    file_stat = stat(filename)
                except OSError as ose:
                    if filename not in no_stat_file:
                        no_stat_file.add(filename)
                        log.warn("DirectoryCheck: could not stat file %s - "
                                 "%s" % (filename, ose))
                else:
                    directory_files += 1
                    directory_bytes += file_stat.st_size

        # number of files
        self.gauge("directory.files_count", directory_files,
                   dimensions=dimensions)
        # total file size
        self.gauge("directory.size_bytes", directory_bytes,
                   dimensions=dimensions)
        log.debug("DirectoryCheck: Directory {0} size {1} bytes with {2} "
                  "files in it.".format(directory_name, directory_bytes,
                                        directory_files))



# (C) Copyright 2015 Hewlett Packard Enterprise Development Company LP
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Classes for detection of running resources to be monitored.

    Detection classes should be platform independent
"""
import ast
import logging
import sys

log = logging.getLogger(__name__)


class Plugin(object):
    """Abstract class for the monasca-agent plugin detection."""

    def __init__(self, template_dir, overwrite=True, args=None):
        self.available = False
        self.template_dir = template_dir
        self.dependencies = ()
        self.overwrite = overwrite
        if args is not None and isinstance(args, str):
            try:
                # Turn 'hostname=host type=ping' to dictionary
                # {'hostname': 'host', 'type': 'ping'}
                self.args = dict([a.split('=') for a in args.split()])
            except Exception:
                log.exception('Error parsing detection arguments')
                sys.exit(1)
        elif isinstance(args, dict):
            self.args = args
        else:
            self.args = None
        self._detect()

    def _detect(self):
        """Run detection.

        Set self.available True if the service is detected.
        """
        raise NotImplementedError

    def build_config(self):
        """Build the config as a Plugins object and return."""
        raise NotImplementedError

    def build_config_with_name(self):
        """Build the config and then add 'built_by' field.

           Build the config and then add 'built_by' field to each instance in
           the config.

           built_by is set to the plugin name

           :return: An agent_config.Plugins object
        """
        conf = self.build_config()
        if conf is None:
            return None
        for plugin_type in conf.values():
            for inst in plugin_type['instances']:
                inst['built_by'] = self.__class__.__name__
        return conf

    def dependencies_installed(self):
        """Return True if dependencies are installed.

        """
        raise NotImplementedError

    @staticmethod
    def literal_eval(testval):
        """Return a literal boolean value if applicable

        """
        if 'false' in str(testval).lower() or 'true' in str(testval).lower():
            return ast.literal_eval(str(testval).capitalize())
        else:
            return testval

    @property
    def name(self):
        """Return _name if set otherwise the class name.

        """
        if '_name' in self.__dict__:
            return self._name
        else:
            return self.__class__.__name__



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import sys
import os
import logging
import unittest

from collector import modules


log = logging.getLogger('datadog.test')

default_target = 'DEFAULT'
specified_target = 'SPECIFIED'
has_been_mutated = False


class TestModuleLoad(unittest.TestCase):

    def setUp(self):
        sys.modules[__name__].has_been_mutated = True
        if 'tests.target_module' in sys.modules:
            del sys.modules['tests.target_module']

    def tearDown(self):
        sys.modules[__name__].has_been_mutated = False

    def test_cached_module(self):
        """Modules already in the cache should be reused"""
        self.assertTrue(modules.load('%s:has_been_mutated' % __name__))

    def test_cache_population(self):
        """Python module cache should be populated"""
        self.assertTrue('tests.target_module' not in sys.modules)
        modules.load('tests.target_module')
        self.assertTrue('tests.target_module' in sys.modules)

    def test_modname_load_default(self):
        """When the specifier contains no module name, any provided default
        should be used"""
        self.assertEqual(
            modules.load(
                'tests.target_module',
                'default_target'),
            'DEFAULT'
        )

    def test_modname_load_specified(self):
        """When the specifier contains a module name, any provided default
        should be overridden"""
        self.assertEqual(
            modules.load(
                'tests.target_module:specified_target',
                'default_target'),
            'SPECIFIED'
        )

    def test_pathname_load_finds_package(self):
        """"Loading modules by absolute path should correctly set the name of
        the loaded module to include any package containing it."""
        m = modules.load(os.getcwd() + '/tests/target_module.py')
        self.assertEqual(m.__name__, 'tests.target_module')



# Copyright 2019 SUSE LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging
import unittest

from monasca_setup.detection.plugins.cert_file_check import CertificateFileCheck

LOG = logging.getLogger('monasca_setup.detection.plugins.cert_check')


class TestCertFileCheck(unittest.TestCase):

    def setUp(self):
        unittest.TestCase.setUp(self)
        with unittest.mock.patch.object(CertificateFileCheck, '_detect') as mock_detect:
            self.cert_obj = CertificateFileCheck('temp_dir')
            self.assertTrue(mock_detect.called)
            self.cert_obj.args = {'cert_files': '/etc/myservice/myserver.pem'}

    def test_detect(self):
        self.cert_obj.available = False
        with unittest.mock.patch.object(self.cert_obj, '_check_required_args',
                          return_value=True) as mock_check_required_args:
            self.cert_obj._detect()
            self.assertTrue(self.cert_obj.available)
            self.assertTrue(mock_check_required_args.called)

    def _build_config(self):
        with unittest.mock.patch.object(self.cert_obj, '_build_instance',
                          return_value={}) as mock_build_instance:
            result = self.cert_obj.build_config()
            self.assertTrue(mock_build_instance.called)
            self.assertEqual(
                result['cert_file_check']['instances'][0]['cert_file'],
                '/etc/myservice/myserver.pem')
            self.assertEqual(result['cert_file_check']['instances'][0]['name'],
                             '/etc/myservice/myserver.pem')
            return result

    def test_build_config_without_args(self):
        result = self._build_config()
        self.assertEqual(
            result['cert_file_check']['init_config']['collect_period'],
            3600)

    def test_build_config_with_args(self):
        self.cert_obj.args.update({'collect_period': 1200})
        result = self._build_config()
        self.assertEqual(
            result['cert_file_check']['init_config']['collect_period'],
            1200)



# (C) Copyright 2016 Hewlett Packard Enterprise Development LP
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import unittest

import monasca_agent.common.metrics as metrics

SAMPLE_RATE = 1


class TestMetrics(unittest.TestCase):
    def test_Gauge(self):
        tenant_name = "test_gauge"
        metric_name = "foo"
        dimensions = {'a': 'b', 'c': 'd'}

        gauge = metrics.Gauge(metric_name, dimensions, tenant_name)

        gauge.sample(0, SAMPLE_RATE, 1)

        envelope = gauge.flush()[0]
        measurement = envelope['measurement']

        self.assertEqual(envelope['tenant_id'], tenant_name)
        self.assertEqual(measurement['name'], metric_name)
        self.assertEqual(measurement['dimensions'], dimensions)
        self.assertEqual(measurement['value'], 0)
        self.assertEqual(measurement['timestamp'], 1000)

        gauge.sample(1, SAMPLE_RATE, 2)

        envelope = gauge.flush()[0]
        measurement = envelope['measurement']

        self.assertEqual(envelope['tenant_id'], tenant_name)
        self.assertEqual(measurement['name'], metric_name)
        self.assertEqual(measurement['dimensions'], dimensions)
        self.assertEqual(measurement['value'], 1)
        self.assertEqual(measurement['timestamp'], 2000)

        gauge.sample(100.5212, SAMPLE_RATE, 125)

        envelope = gauge.flush()[0]
        measurement = envelope['measurement']

        self.assertEqual(envelope['tenant_id'], tenant_name)
        self.assertEqual(measurement['name'], metric_name)
        self.assertEqual(measurement['dimensions'], dimensions)
        self.assertEqual(measurement['value'], 100.5212)
        self.assertEqual(measurement['timestamp'], 125000)

        results = gauge.flush()
        self.assertEqual(results, [])

    def test_Counter(self):
        tenant_name = "test_counter"
        metric_name = "bar"
        dimensions = {'a': 1, 'c': 2}

        counter = metrics.Counter(metric_name, dimensions, tenant_name)

        # single counter value
        counter.sample(5, SAMPLE_RATE, 1)

        envelope = counter.flush()[0]
        measurement = envelope['measurement']
        self.assertEqual(envelope['tenant_id'], tenant_name)
        self.assertEqual(measurement['name'], metric_name)
        self.assertEqual(measurement['dimensions'], dimensions)
        self.assertEqual(measurement['value'], 5)
        self.assertEqual(measurement['timestamp'], 1000)

        # multiple counter value with different timestamps: add
        counter.sample(5, SAMPLE_RATE, 1)
        counter.sample(6, SAMPLE_RATE, 2)

        envelope = counter.flush()[0]
        measurement = envelope['measurement']
        self.assertEqual(envelope['tenant_id'], tenant_name)
        self.assertEqual(measurement['name'], metric_name)
        self.assertEqual(measurement['dimensions'], dimensions)
        self.assertEqual(measurement['value'], 11)
        self.assertEqual(measurement['timestamp'], 2000)

        # multiple counter values with same timestamp: add
        counter.sample(5, SAMPLE_RATE, 3)
        counter.sample(5, SAMPLE_RATE, 3)

        envelope = counter.flush()[0]
        measurement = envelope['measurement']
        self.assertEqual(envelope['tenant_id'], tenant_name)
        self.assertEqual(measurement['name'], metric_name)
        self.assertEqual(measurement['dimensions'], dimensions)
        self.assertEqual(measurement['value'], 5+5)
        self.assertEqual(measurement['timestamp'], 3000)

        # Invalid metric values: ignore
        counter.sample("WEGONI", SAMPLE_RATE, 2)
        results = counter.flush()
        self.assertEqual(results, [])

    def test_Rate(self):
        tenant_name = "test_rate"
        metric_name = "baz"
        dimensions = {'a': 2, 'c': 3}

        rate = metrics.Rate(metric_name, dimensions, tenant_name)

        # single sample without predecessor: no rate can be calculated
        rate.sample(5, SAMPLE_RATE, 1)
        self.assertEqual(rate.flush(), [])

        # zero difference between samples: rate 0
        rate.sample(5, SAMPLE_RATE, 2)

        envelope = rate.flush()[0]
        measurement = envelope['measurement']
        self.assertEqual(envelope['tenant_id'], tenant_name)
        self.assertEqual(measurement['name'], metric_name)
        self.assertEqual(measurement['dimensions'], dimensions)
        self.assertEqual(measurement['value'], 0.0)
        self.assertEqual(measurement['timestamp'], 2000)

        # samples (5,10) in 1 sec interval: rate 5/sec.
        rate.sample(10, SAMPLE_RATE, 3)

        envelope = rate.flush()[0]
        measurement = envelope['measurement']
        self.assertEqual(envelope['tenant_id'], tenant_name)
        self.assertEqual(measurement['name'], metric_name)
        self.assertEqual(measurement['dimensions'], dimensions)
        self.assertEqual(measurement['value'], 5)
        self.assertEqual(measurement['timestamp'], 3000)

        # conflicting values for same timestamp: no result, but keep last sample for next rate calc.
        rate.sample(12, SAMPLE_RATE, 3)
        self.assertEqual(rate.flush(), [])

        # zero difference between samples, incomplete previous interval T: rate 0/sec.
        rate.sample(12, SAMPLE_RATE, 4)

        envelope = rate.flush()[0]
        measurement = envelope['measurement']
        self.assertEqual(envelope['tenant_id'], tenant_name)
        self.assertEqual(measurement['name'], metric_name)
        self.assertEqual(measurement['dimensions'], dimensions)
        self.assertEqual(measurement['value'], 0.0)
        self.assertEqual(measurement['timestamp'], 4000)

        # several samples (13, 14) in interval, take last values of T1 and T0 for rate calc: rate = (14-12)/(6-4)
        rate.sample(13, SAMPLE_RATE, 5)
        rate.sample(14, SAMPLE_RATE, 6)

        envelope = rate.flush()[0]
        measurement = envelope['measurement']
        self.assertEqual(envelope['tenant_id'], tenant_name)
        self.assertEqual(measurement['name'], metric_name)
        self.assertEqual(measurement['dimensions'], dimensions)
        self.assertEqual(measurement['value'], 1)
        self.assertEqual(measurement['timestamp'], 6000)

        # negative rate: often result of a restart, but that should not be hidden
        rate.sample(1, SAMPLE_RATE, 7)

        envelope = rate.flush()[0]
        measurement = envelope['measurement']
        self.assertEqual(envelope['tenant_id'], tenant_name)
        self.assertEqual(measurement['name'], metric_name)
        self.assertEqual(measurement['dimensions'], dimensions)
        self.assertEqual(measurement['value'], -13)
        self.assertEqual(measurement['timestamp'], 7000)

        # recover from negative rate
        rate.sample(2, SAMPLE_RATE, 8)

        envelope = rate.flush()[0]
        measurement = envelope['measurement']
        self.assertEqual(envelope['tenant_id'], tenant_name)
        self.assertEqual(measurement['name'], metric_name)
        self.assertEqual(measurement['dimensions'], dimensions)
        self.assertEqual(measurement['value'], 1)
        self.assertEqual(measurement['timestamp'], 8000)



# (C) Copyright 2016 Hewlett Packard Enterprise Development Company LP
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import configparser
import logging
import os
import psutil
import re
from unittest import mock
from unittest import TestCase

from oslo_config import cfg

from monasca_setup.detection import utils
from monasca_setup.detection.plugins.ovs import Ovs


LOG = logging.getLogger('monasca_setup.detection.plugins.ovs')


class ps_util_get_proc:
    cmdLine = ['/etc/neutron/neutron.conf']
    detect_warning = False
    def as_dict(self, attrs=None):
        return {'name': 'neutron-openvsw',
                'cmdline': self.cmdline()}

    def cmdline(self):
        if not ps_util_get_proc.detect_warning:
            return ['neutron-openvsw',
                    ps_util_get_proc.cmdLine[0]]
        else:
            return ['/opt/fake.txt']


class TestOvs(TestCase):

    def setUp(self):
        TestCase.setUp(self)
        with mock.patch.object(Ovs, '_detect') as mock_detect:
            self.ovs_obj = Ovs('temp_dir')
            self.has_option = [True, False, True, False, False, True]
            self.get_value = [mock.MagicMock(), mock.MagicMock(), mock.MagicMock(),
                              mock.MagicMock(), mock.MagicMock(), 'http://10.10.10.10',
                              'region1']
            self.assertTrue(mock_detect.called)

    def _detect(self, ovs_obj, file_config_valid=True):
        ovs_obj.neutron_conf = None
        ovs_obj.available = False

        with mock.patch.object(cfg, 'CONF') as mock_conf, \
                mock.patch.object(psutil, 'process_iter',
                                  return_value=[ps_util_get_proc()]) \
                        as mock_process_iter, \
                mock.patch.object(os.path, 'isfile', return_value=True) \
                        as mock_isfile,\
                mock.patch.object(ovs_obj, 'dependencies_installed', return_value=True) \
                        as dependencies,\
                mock.patch.object(ovs_obj, '_is_neutron_conf_valid',
                                  return_value=file_config_valid) as _:
            ovs_obj._detect()
            self.assertTrue(mock_process_iter.called)
            if not ps_util_get_proc.cmdLine:
                self.assertFalse(mock_isfile.called)

    def _build_config(self, ovs_obj, dependencies_installed=True):
        ovs_obj.conf = mock.MagicMock()
        ovs_obj.conf.default_config_files = ovs_obj.neutron_conf
        ovs_obj.conf.default_config_dirs = os.path.abspath(os.path.join(ovs_obj.neutron_conf, os.pardir))
        with mock.patch.object(configparser, 'SafeConfigParser') as mock_config_parser:
            config_parser_obj = mock_config_parser.return_value

            with mock.patch.object(cfg, 'CONF') as mock_conf, \
                    mock.patch.object(LOG, 'info') as mock_log_info, \
                    mock.patch.object(ovs_obj, 'has_option',
                                      side_effect=self.has_option) as mock_has_option, \
                    mock.patch.object(ovs_obj, 'get_option',
                                      side_effect=self.get_value) as mock_get:
                result = ovs_obj.build_config()
                if dependencies_installed:
                    self.assertTrue(mock_log_info.called)
                    self.assertTrue(mock_has_option.called)
                    self.assertTrue(mock_get.called)
                    if not self.has_option[-1]:
                        self.assertIn(str(('nova', 'region_name')),
                                      str(mock_get.call_args_list[-1]))
                    else:
                        self.assertIn(str(('service_auth', 'region')),
                                      str(mock_get.call_args_list[-1]))
                return result

    def _build_config_with_arg(self, ovs_obj):
            result = self._build_config(ovs_obj)
            self.assertEqual(result['ovs']['init_config']['neutron_refresh'],
                             13000)
            self.assertFalse(result['ovs']['init_config']['network_use_bits'])
            self.assertIsInstance(result['ovs']['init_config']['username'],
                                  mock.MagicMock)
            self.assertIsInstance(result['ovs']['init_config']['password'],
                                  mock.MagicMock)
            self.assertIsInstance(result['ovs']['init_config']['project_name'],
                                  mock.MagicMock)
            self.assertEqual(result['ovs']['init_config']['auth_url'],
                             'http://10.10.10.10')
            self.assertEqual(result['ovs']['init_config']['region_name'],
                             'region1')
            self.assertEqual(result['ovs']['init_config']['cache_dir'],
                             "/dev/shm")
            self.assertEqual(result['ovs']['init_config']['ovs_cmd'],
                             "sudo /usr/bin/ovs-vsctl")
            self.assertFalse(result['ovs']['init_config']['use_absolute_metrics'])
            self.assertTrue(result['ovs']['init_config']['use_rate_metrics'])
            self.assertTrue(result['ovs']['init_config']['use_health_metrics'])
            return result

    def _build_config_without_args(self, ovs_obj):
        result = self._build_config(ovs_obj)
        self.assertEqual(result['ovs']['init_config']['neutron_refresh'],
                         14400)
        self.assertFalse(result['ovs']['init_config']['network_use_bits'])
        self.assertEqual(result['ovs']['init_config']['cache_dir'],
                         "/dev/shm")
        self.assertEqual(result['ovs']['init_config']['ovs_cmd'],
                         "sudo /usr/bin/ovs-vsctl")
        self.assertEqual(result['ovs']['init_config']['included_interface_re'],
                         'qg.*|vhu.*|sg.*')
        self.assertIsInstance(result['ovs']['init_config']['username'],
                              mock.MagicMock)
        self.assertIsInstance(result['ovs']['init_config']['password'],
                              mock.MagicMock)
        self.assertIsInstance(result['ovs']['init_config']['project_name'],
                              mock.MagicMock)
        self.assertTrue(result['ovs']['init_config']['use_absolute_metrics'])
        self.assertTrue(result['ovs']['init_config']['use_rate_metrics'])
        self.assertTrue(result['ovs']['init_config']['use_health_metrics'])
        return result

    def test_detect(self):
        with mock.patch.object(utils, 'load_oslo_configuration'):
            self._detect(self.ovs_obj)
            self.assertTrue(self.ovs_obj.available)
            self.assertEqual(self.ovs_obj.neutron_conf,
                             '/etc/neutron/neutron.conf')

    def test_detect_invalid_config_file(self):
        self._detect(self.ovs_obj, file_config_valid=False)
        self.assertFalse(self.ovs_obj.available)
        self.assertIsNone(self.ovs_obj.neutron_conf)

    def test_detect_devstack(self):
        ps_util_get_proc.cmdLine = ['--config-file=/opt/stack/neutron.conf']
        with mock.patch.object(utils, 'load_oslo_configuration'):
            self._detect(self.ovs_obj)
            self.assertTrue(self.ovs_obj.available)
            self.assertEqual(self.ovs_obj.neutron_conf, '/opt/stack/neutron.conf')

    def test_detect_info(self):
        with mock.patch.object(LOG, 'info') as mock_log_warn:
            ps_util_get_proc.detect_warning = True
            self._detect(self.ovs_obj)
            self.assertFalse(self.ovs_obj.available)
            self.assertIsNone(self.ovs_obj.neutron_conf)
            self.assertTrue(mock_log_warn.called)

    def test_detect_conf_file_path_given(self):
        self.ovs_obj.neutron_conf = None
        self.ovs_obj.args = {'conf_file_path': '/opt/stack/neutron.conf'}

        with mock.patch.object(utils, 'load_oslo_configuration') as mock_conf, \
                mock.patch.object(psutil, 'process_iter',
                                  return_value=[ps_util_get_proc()]) as mock_process_iter, \
                mock.patch.object(os.path, 'isfile', return_value=True) as mock_isfile, \
                mock.patch.object(self.ovs_obj, 'dependencies_installed',
                                  return_value=True) as dependencies, \
                mock.patch.object(self.ovs_obj, '_is_neutron_conf_valid',
                                  return_value=True) as _:
            self.ovs_obj._detect()
            self.assertTrue(mock_isfile.called)
            self.assertTrue(self.ovs_obj.available)
            self.assertEqual(self.ovs_obj.neutron_conf,
                             '/opt/stack/neutron.conf')

    def test_build_config(self):
        self.ovs_obj.neutron_conf = 'neutron-conf'
        self._build_config_without_args(self.ovs_obj)

    def test_build_config_with_args(self):
        with mock.patch.object(LOG, 'warn') as mock_log_warn:
            self.ovs_obj.neutron_conf = 'neutron-conf'
            self.ovs_obj.args = {'username': 'admin',
                                 'password': 'password',
                                 'project_name': 'tenant',
                                 'auth_url': '10.10.10.20',
                                 'region_name': 'region0',
                                 'neutron_refresh': 13000,
                                 'use_absolute_metrics': False}
            result = self._build_config_with_arg(self.ovs_obj)
            self.assertTrue(mock_log_warn.called)
            self.assertEqual(mock_log_warn.call_count, 5)
            self.assertEqual(result['ovs']['init_config']['included_interface_re'],
                             'qg.*|vhu.*|sg.*')

    def test_dependencies_not_installed(self):
        result = self.ovs_obj.dependencies_installed()
        self.assertEqual(result, False)

    def test_build_config_invalid_arg_warning(self):
        with mock.patch.object(LOG, 'warn') as mock_log_warn:
            self.ovs_obj.neutron_conf = 'neutron-conf'
            self.ovs_obj.args = {'username': 'admin',
                                 'password': 'password',
                                 'project_name': 'tenant',
                                 'user_domain_name': 'default',
                                 'project_domain_name': 'default',
                                 'auth_url': '10.10.10.20',
                                 'region_name': 'region0',
                                 'neutron_refresh': 13000,
                                 'use_absolute_metrics': False,
                                 'invalid_arg': 'inv-arg'}
            result = self._build_config_with_arg(self.ovs_obj)
            self.assertTrue(mock_log_warn.called)
            self.assertEqual(mock_log_warn.call_count, 8)
            self.assertEqual(result['ovs']['init_config']['included_interface_re'],
                             'qg.*|vhu.*|sg.*')

    def test_build_config_if_auth_version(self):
        self.ovs_obj.neutron_conf = 'neutron-conf'
        self.has_option = [True, False, True, False, True, True]
        self.get_value = [mock.MagicMock(), mock.MagicMock(), mock.MagicMock(), mock.MagicMock(),
                          mock.MagicMock(), 'http://10.10.10.10',
                          'http://10.10.10.10', 'region1']
        result = self._build_config_without_args(self.ovs_obj)
        self.assertEqual(result['ovs']['init_config']['auth_url'],
                         'http://10.10.10.10')

    def test_build_config_if_auth_url_has_version(self):
        self.ovs_obj.neutron_conf = 'neutron-conf'
        self.has_option = [True, False, True, False, True, True]
        self.get_value = [mock.MagicMock(), mock.MagicMock(), mock.MagicMock(), mock.MagicMock(),
                          mock.MagicMock(), 'http://10.10.10.10/v1',
                          'http://10.10.10.10/v1', 'region1']
        result = self._build_config_without_args(self.ovs_obj)
        self.assertEqual(result['ovs']['init_config']['auth_url'],
                         'http://10.10.10.10/v1')

    def test_build_config_region_name_from_nova(self):
        self.ovs_obj.neutron_conf = 'neutron-conf'
        self.has_option = [True, False, True, False, False, False]
        self.get_value = [mock.MagicMock(), mock.MagicMock(), mock.MagicMock(), mock.MagicMock(),
                          mock.MagicMock(), 'http://10.10.10.10', 'region2']
        result = self._build_config_without_args(self.ovs_obj)
        self.assertEqual(result['ovs']['init_config']['auth_url'],
                         'http://10.10.10.10')
        self.assertEqual(result['ovs']['init_config']['region_name'],
                         'region2')

    def test_build_config_with_valid_interface_re(self):
        self.ovs_obj.neutron_conf = 'neutron-conf'
        self.ovs_obj.args = {'included_interface_re': 'tap.*',
                             'neutron_refresh': 13000,
                             'use_absolute_metrics': False}
        result = self._build_config_with_arg(self.ovs_obj)
        self.assertEqual(result['ovs']['init_config']['included_interface_re'],
                         'tap.*')

    def test_build_config_with_invalid_interface_re(self):
        self.ovs_obj.neutron_conf = 'neutron-conf'
        self.ovs_obj.args = {'included_interface_re': '[',
                             'neutron_refresh': 13000}

        with mock.patch.object(re, 'compile', side_effect=re.error('error')) as mock_re_error, \
                mock.patch.object(LOG, 'exception') as mock_log:
            self.assertRaises(Exception, self._build_config_with_arg, self.ovs_obj)
            self.assertTrue(mock_re_error.called)
            self.assertTrue(mock_log.called)



# (C) Copyright 2017,2018 Hewlett Packard Enterprise Development LP
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging
import requests

from monasca_agent.collector import checks
from monasca_agent.collector.checks import utils
from monasca_agent.common.util import rollup_dictionaries

log = logging.getLogger(__name__)

DEFAULT_TIMEOUT = 5
DEFAULT_KUBELET_PORT = "10255"
DEFAULT_CADVISOR_PORT = "4194"
CADVISOR_METRIC_URL = "/api/v2.0/stats?type=docker&recursive=true&count=1"
CADVISOR_SPEC_URL = "/api/v2.0/spec?type=docker&recursive=true"
POD_PHASE = {"Succeeded": 0,
             "Running": 1,
             "Pending": 2,
             "Failed": 3,
             "Unknown": 4}
REPORT_CONTAINER_METRICS = False
CADVISOR_METRICS = {
    "cpu_metrics": {
        "system": "cpu.system_time",
        "total": "cpu.total_time",
        "user": "cpu.user_time"
    },
    "memory_metrics": {
        "rss": "mem.rss_bytes",
        "swap": "mem.swap_bytes",
        "cache": "mem.cache_bytes",
        "usage": "mem.used_bytes",
        "failcnt": "mem.fail_count",
        "working_set": "mem.working_set_bytes"
    },
    "filesystem_metrics": {
        "capacity": "fs.total_bytes",
        "usage": "fs.usage_bytes",
        "writes_completed": "fs.writes",
        "reads_completes": "fs.reads",
        "io_in_progress": "fs.io_current"
    },
    "network_metrics": {
        "rx_bytes": "net.in_bytes",
        "tx_bytes": "net.out_bytes",
        "rx_packets": "net.in_packets",
        "tx_packets": "net.out_packets",
        "rx_dropped": "net.in_dropped_packets",
        "tx_dropped": "net.out_dropped_packets",
        "rx_errors": "net.in_errors",
        "tx_errors": "net.out_errors",
    },
    "diskio_metrics": {
        "Read": "io.read_bytes",
        "Write": "io.write_bytes"
    }
}

# format: (cadvisor metric name, [metric types], [metric units])
METRIC_TYPES_UNITS = {
    "cpu.system_time": (["gauge", "rate"], ["core_seconds", "cores_seconds_per_second"]),
    "cpu.total_time": (["gauge", "rate"], ["core_seconds", "cores_seconds_per_second"]),
    "cpu.user_time": (["gauge", "rate"], ["core_seconds", "cores_seconds_per_second"]),
    "mem.rss_bytes": (["gauge"], ["bytes"]),
    "mem.swap_bytes": (["gauge"], ["bytes"]),
    "mem.cache_bytes": (["gauge"], ["bytes"]),
    "mem.used_bytes": (["gauge"], ["bytes"]),
    "mem.fail_count": (["gauge"], ["count"]),
    "mem.working_set_bytes": (["gauge"], ["bytes"]),
    "fs.total_bytes": (["gauge"], ["bytes"]),
    "fs.usage_bytes": (["gauge"], ["bytes"]),
    "fs.writes": (["gauge", "rate"], ["bytes", "bytes_per_second"]),
    "fs.reads": (["gauge", "rate"], ["bytes", "bytes_per_second"]),
    "fs.io_current": (["gauge"], ["bytes"]),
    "net.in_bytes": (["gauge", "rate"], ["total_bytes", "total_bytes_per_second"]),
    "net.out_bytes": (["gauge", "rate"], ["total_bytes", "total_bytes_per_second"]),
    "net.in_packets": (["gauge", "rate"], ["total_packets", "total_packets_per_second"]),
    "net.out_packets": (["gauge", "rate"], ["total_packets", "total_packets_per_second"]),
    "net.in_dropped_packets": (["gauge", "rate"], ["total_packets", "total_packets_per_second"]),
    "net.out_dropped_packets": (["gauge", "rate"], ["total_packets", "total_packets_per_second"]),
    "net.in_errors": (["gauge", "rate"], ["total_errors", "total_errors_per_second"]),
    "net.out_errors": (["gauge", "rate"], ["total_errors", "total_errors_per_second"]),
    "io.read_bytes": (["gauge", "rate"], ["total_bytes", "total_bytes_per_second"]),
    "io.write_bytes": ("write_bytes", ["gauge", "rate"], ["total_bytes", "total_bytes_per_second"])
}


class Kubernetes(checks.AgentCheck):
    """Queries Kubelet for metadata/health data and then cAdvisor for container metrics.
    """
    def __init__(self, name, init_config, agent_config, instances=None):
        checks.AgentCheck.__init__(self, name, init_config, agent_config, instances)
        if instances is not None and len(instances) > 1:
            raise Exception('Kubernetes check only supports one configured instance.')
        self.connection_timeout = int(init_config.get('connection_timeout', DEFAULT_TIMEOUT))
        self.host = None
        self.report_container_metrics = init_config.get(
            'report_container_metrics', REPORT_CONTAINER_METRICS)
        self.report_container_mem_percent = init_config.get('report_container_mem_percent', True)
        self.kubernetes_connector = None

    def prepare_run(self):
        """Set up Kubernetes connection information"""
        instance = self.instances[0]
        self.host = instance.get("host", None)
        derive_host = instance.get("derive_host", False)
        if not self.host:
            if derive_host:
                self.kubernetes_connector = utils.KubernetesConnector(self.connection_timeout)
                self.host = self.kubernetes_connector.get_agent_pod_host()
            else:
                exception_message = "Either host or derive host must be set when " \
                                    "running Kubernetes plugin."
                self.log.exception(exception_message)
                raise Exception(exception_message)

    def check(self, instance):
        cadvisor, kubelet = self._get_urls(instance)
        kubernetes_labels = instance.get('kubernetes_labels', ["app"])
        container_dimension_map = {}
        pod_dimensions_map = {}
        memory_limit_map = {}
        dimensions = self._set_dimensions(None, instance)
        # Remove hostname from dimensions as the majority of the metrics are not
        # tied to the hostname.
        del dimensions['hostname']
        kubelet_health_status = self._get_api_health("{}/healthz".format(kubelet))
        self.gauge(
            "kubelet.health_status",
            0 if kubelet_health_status else 1,
            dimensions=dimensions)
        try:
            pods = self._get_result("{}/pods".format(kubelet))
        except Exception as e:
            self.log.exception("Error getting data from kubelet - {}".format(e))
        else:
            self._process_pods(pods['items'],
                               kubernetes_labels,
                               dimensions,
                               container_dimension_map,
                               pod_dimensions_map,
                               memory_limit_map)
            self._process_containers(cadvisor,
                                     dimensions,
                                     container_dimension_map,
                                     pod_dimensions_map,
                                     memory_limit_map)

    def _get_urls(self, instance):
        base_url = "http://{}".format(self.host)
        cadvisor_port = instance.get('cadvisor_port', DEFAULT_CADVISOR_PORT)
        kubelet_port = instance.get('kubelet_port', DEFAULT_KUBELET_PORT)
        cadvisor_url = "{}:{}".format(base_url, cadvisor_port)
        kubelet_url = "{}:{}".format(base_url, kubelet_port)
        return cadvisor_url, kubelet_url

    def _get_result(self, request_url, as_json=True):
        result = requests.get(request_url, timeout=self.connection_timeout)
        return result.json() if as_json else result

    def _get_api_health(self, health_url):
        try:
            result = self._get_result(health_url, as_json=False)
        except Exception as e:
            self.log.error(
                "Error connecting to the health endpoint {} with exception {}".format(
                    health_url, e))
            return False
        else:
            api_health = False
            for line in result.iter_lines():
                if line == 'ok':
                    api_health = True
                    break
            return api_health

    def _process_pods(
            self,
            pods,
            kubernetes_labels,
            dimensions,
            container_dimension_map,
            pod_dimensions_map,
            memory_limit_map):
        for pod in pods:
            pod_status = pod['status']
            pod_spec = pod['spec']
            pod_containers = pod_spec.get('containers', None)
            container_statuses = pod_status.get('containerStatuses', None)
            if not pod_containers or not container_statuses:
                # Pod does not have any containers assigned to it no-op going to next pod
                continue
            pod_dimensions = dimensions.copy()
            pod_dimensions.update(
                utils.get_pod_dimensions(
                    self.kubernetes_connector,
                    pod['metadata'],
                    kubernetes_labels))
            pod_key = pod_dimensions['pod_name'] + pod_dimensions['namespace']
            pod_dimensions_map[pod_key] = pod_dimensions
            pod_retry_count = 0

            name2id = {}

            for container_status in container_statuses:
                container_restart_count = container_status['restartCount']
                container_dimensions = pod_dimensions.copy()
                container_name = container_status['name']
                container_dimensions['container_name'] = container_name
                container_dimensions['image'] = container_status['image']
                container_id = container_status.get('containerID', '').split('//')[-1]
                name2id[container_name] = container_id
                container_dimension_map[container_id] = container_dimensions
                if self.report_container_metrics:
                    container_ready = 0 if container_status['ready'] else 1
                    self.gauge(
                        "container.ready_status",
                        container_ready,
                        container_dimensions,
                        hostname="SUPPRESS")
                    self.gauge(
                        "container.restart_count",
                        container_restart_count,
                        container_dimensions,
                        hostname="SUPPRESS")
                # getting an aggregated value for pod restart count
                pod_retry_count += container_restart_count

            # Report limit/request metrics
            if self.report_container_metrics or self.report_container_mem_percent:
                self._report_container_limits(
                    pod_containers, container_dimension_map, name2id, memory_limit_map)

            self.gauge("pod.restart_count", pod_retry_count, pod_dimensions, hostname="SUPPRESS")
            self.gauge(
                "pod.phase",
                POD_PHASE.get(
                    pod_status['phase']),
                pod_dimensions,
                hostname="SUPPRESS")

    def _report_container_limits(
            self,
            pod_containers,
            container_dimension_map,
            name2id,
            memory_limit_map):
        for container in pod_containers:
            container_name = container['name']
            container_dimensions = container_dimension_map[name2id[container_name]]
            if 'resources' not in container:
                self.log.debug(
                    "Container %s does not have limits or requests set", container_name)
                continue
            container_resources = container['resources']
            if 'limits' not in container_resources:
                self.log.debug("Container %s does not have limits set", container_name)
            else:
                container_limits = container_resources['limits']
                if 'cpu' in container_limits:
                    cpu_limit = container_limits['cpu']
                    cpu_value = self._convert_cpu_to_cores(cpu_limit)
                    if self.report_container_metrics:
                        self.gauge(
                            "container.cpu.limit",
                            cpu_value,
                            container_dimensions,
                            hostname="SUPPRESS")
                else:
                    self.log.debug(
                        "Container %s does not have cpu limit set", container_name)
                if 'memory' in container_limits:
                    memory_limit = container_limits['memory']
                    memory_in_bytes = utils.convert_memory_string_to_bytes(memory_limit)
                    if self.report_container_metrics:
                        self.gauge(
                            "container.memory.limit_bytes",
                            memory_in_bytes,
                            container_dimensions,
                            hostname="SUPPRESS")
                    if self.report_container_mem_percent:
                        container_key = container_name + " " + container_dimensions["namespace"]
                        if container_key not in memory_limit_map:
                            memory_limit_map[container_key] = memory_in_bytes
                else:
                    self.log.debug(
                        "Container %s does not have memory limit set", container_name)
            if 'requests' not in container_resources:
                self.log.debug("Container %s does not have requests set", container_name)
            else:
                container_requests = container_resources['requests']
                if 'cpu' in container_requests:
                    cpu_request = container_requests['cpu']
                    cpu_value = self._convert_cpu_to_cores(cpu_request)
                    if self.report_container_metrics:
                        self.gauge(
                            "container.request.cpu",
                            cpu_value,
                            container_dimensions,
                            hostname="SUPPRESS")
                else:
                    self.log.debug(
                        "Container %s does not have cpu request set", container_name)
                if 'memory' in container_requests:
                    memory_request = container_requests['memory']
                    memory_in_bytes = utils.convert_memory_string_to_bytes(memory_request)
                    if self.report_container_metrics:
                        self.gauge(
                            "container.request.memory_bytes",
                            memory_in_bytes,
                            container_dimensions,
                            hostname="SUPPRESS")
                else:
                    self.log.debug(
                        "Container %s does not have memory request set", container_name)

    def _convert_cpu_to_cores(self, cpu_string):
        """Kubernetes reports cores in millicores in some instances.
        This method makes sure when we report on cpu they are all in cores
        """
        if "m" in cpu_string:
            cpu = float(cpu_string.split('m')[0])
            return cpu / 1000
        return float(cpu_string)

    def _send_metrics(self, metric_name, value, dimensions, metric_types,
                      metric_units):
        for metric_type in metric_types:
            if metric_type == 'rate':
                dimensions.update({'unit': metric_units[
                    metric_types.index('rate')]})
                self.rate(metric_name + "_sec", value, dimensions,
                          hostname="SUPPRESS" if "pod_name" in dimensions else None)
            elif metric_type == 'gauge':
                dimensions.update({'unit': metric_units[
                    metric_types.index('gauge')]})
                self.gauge(metric_name, value, dimensions,
                           hostname="SUPPRESS" if "pod_name" in dimensions else None)

    def _parse_memory(self, memory_data, container_dimensions, pod_key, pod_map, memory_limit_map):
        memory_metrics = CADVISOR_METRICS['memory_metrics']
        for cadvisor_key, metric_name in memory_metrics.items():
            if cadvisor_key in memory_data:
                metric_value = memory_data[cadvisor_key]
                if self.report_container_metrics:
                    self._send_metrics("container." + metric_name,
                                       metric_value,
                                       container_dimensions,
                                       METRIC_TYPES_UNITS[metric_name][0],
                                       METRIC_TYPES_UNITS[metric_name][1])
                self._add_pod_metric(metric_name, metric_value, pod_key, pod_map)
                if self.report_container_mem_percent and cadvisor_key == "working_set":
                    if "container_name" in container_dimensions \
                            and "namespace" in container_dimensions:
                        container_key = container_dimensions["container_name"] + \
                            " " + container_dimensions["namespace"]
                        if container_key not in memory_limit_map:
                            continue
                        memory_limit = memory_limit_map[container_key]
                        memory_usage = metric_value
                        memory_usage_percent = (memory_usage / memory_limit) * 100
                        self.gauge(
                            "container.mem.usage_percent",
                            memory_usage_percent,
                            container_dimensions,
                            hostname="SUPPRESS")

    def _parse_filesystem(self, filesystem_data, container_dimensions):
        if not self.report_container_metrics:
            return
        filesystem_metrics = CADVISOR_METRICS['filesystem_metrics']
        for filesystem in filesystem_data:
            file_dimensions = container_dimensions.copy()
            file_dimensions['device'] = filesystem['device']
            for cadvisor_key, metric_name in filesystem_metrics.items():
                if cadvisor_key in filesystem:
                    self._send_metrics("container." + metric_name,
                                       filesystem[cadvisor_key],
                                       file_dimensions,
                                       METRIC_TYPES_UNITS[metric_name][0],
                                       METRIC_TYPES_UNITS[metric_name][1])

    def _parse_network(self, network_data, container_dimensions, pod_key, pod_metrics):
        if 'interfaces' not in network_data:
            return
        network_interfaces = network_data['interfaces']
        network_metrics = CADVISOR_METRICS['network_metrics']
        sum_network_interfaces = {}
        for interface in network_interfaces:
            sum_network_interfaces = rollup_dictionaries(
                sum_network_interfaces, interface)
        for cadvisor_key, metric_name in network_metrics.items():
            if cadvisor_key in sum_network_interfaces:
                metric_value = sum_network_interfaces[cadvisor_key]
                if self.report_container_metrics:
                    self._send_metrics("container." + metric_name,
                                       metric_value,
                                       container_dimensions,
                                       METRIC_TYPES_UNITS[metric_name][0],
                                       METRIC_TYPES_UNITS[metric_name][1])
                self._add_pod_metric(metric_name, metric_value, pod_key,
                                     pod_metrics)

    def _parse_cpu(self, cpu_data, container_dimensions, pod_key, pod_metrics):
        cpu_metrics = CADVISOR_METRICS['cpu_metrics']
        cpu_usage = cpu_data['usage']
        for cadvisor_key, metric_name in cpu_metrics.items():
            if cadvisor_key in cpu_usage:
                # convert nanoseconds to seconds
                cpu_usage_sec = cpu_usage[cadvisor_key] / 1000000000
                if self.report_container_metrics:
                    self._send_metrics("container." + metric_name,
                                       cpu_usage_sec,
                                       container_dimensions,
                                       METRIC_TYPES_UNITS[metric_name][0],
                                       METRIC_TYPES_UNITS[metric_name][1])
                self._add_pod_metric(metric_name, cpu_usage_sec, pod_key, pod_metrics)

    def _parse_diskio(self, diskio_data, container_dimensions, pod_key, pod_metrics):
        diskio_services = diskio_data['io_service_bytes']
        diskio_metrics = CADVISOR_METRICS['diskio_metrics']
        sum_diskio_devices = {}
        for disk_device in diskio_services:
            sum_diskio_devices = rollup_dictionaries(
                sum_diskio_devices, disk_device['stats'])

        for cadvisor_key, metric_name in diskio_metrics.items():
            if cadvisor_key in sum_diskio_devices:
                metric_value = sum_diskio_devices[cadvisor_key]
                if self.report_container_metrics:
                    self._send_metrics("container." + metric_name,
                                       metric_value,
                                       container_dimensions,
                                       METRIC_TYPES_UNITS[metric_name][0],
                                       METRIC_TYPES_UNITS[metric_name][1])
                self._add_pod_metric(metric_name, metric_value, pod_key,
                                     pod_metrics)

    def _add_pod_metric(self, metric_name, metric_value, pod_key, pod_metrics):
        if pod_key:
            if pod_key not in pod_metrics:
                pod_metrics[pod_key] = {}
            if metric_name not in pod_metrics[pod_key]:
                pod_metrics[pod_key][metric_name] = metric_value
            else:
                pod_metrics[pod_key][metric_name] += metric_value

    def _get_container_dimensions(
            self,
            container,
            instance_dimensions,
            container_spec,
            container_dimension_map,
            pod_dimension_map):
        container_id = ""
        # meant to key through pod metrics/dimension dictionaries

        for alias in container_spec["aliases"]:
            if alias in container:
                container_id = alias
                break
        if container_id in container_dimension_map:
            container_dimensions = container_dimension_map[container_id]
            pod_key = container_dimensions['pod_name'] + container_dimensions['namespace']
            return pod_key, container_dimensions
        else:
            container_dimensions = instance_dimensions.copy()
            # Container image being used
            container_dimensions['image'] = container_spec['image']
            # First entry in aliases is container name
            container_dimensions['container_name'] = container_spec['aliases'][0]
            # check if container is a pause container running under a pod. Owns network namespace
            pod_key = None
            if 'labels' in container_spec:
                container_labels = container_spec['labels']
                if 'io.kubernetes.pod.namespace' in container_labels \
                        and 'io.kubernetes.pod.name' in container_labels:
                    pod_key = container_labels['io.kubernetes.pod.name'] + \
                        container_labels['io.kubernetes.pod.namespace']
                    # In case new pods showed up since we got our pod list from the kubelet
                    if pod_key in pod_dimension_map:
                        container_dimensions.update(pod_dimension_map[pod_key])
                        container_dimensions['container_name'] = \
                            container_labels['io.kubernetes.container.name']
                    else:
                        pod_key = None
            return pod_key, container_dimensions

    def _process_containers(
            self,
            cadvisor_url,
            dimensions,
            container_dimension_map,
            pod_dimension_map,
            memory_limit_map):
        try:
            cadvisor_spec_url = cadvisor_url + CADVISOR_SPEC_URL
            cadvisor_metric_url = cadvisor_url + CADVISOR_METRIC_URL
            containers_spec = self._get_result(cadvisor_spec_url)
            containers_metrics = self._get_result(cadvisor_metric_url)
        except Exception as e:
            self.log.error("Error getting data from cadvisor - {}".format(e))
            return
        pod_metrics = {}
        for container, cadvisor_metrics in containers_metrics.items():
            pod_key, container_dimensions = self._get_container_dimensions(
                container,
                dimensions,
                containers_spec[container],
                container_dimension_map,
                pod_dimension_map)
            # Grab first set of metrics from return data
            cadvisor_metrics = cadvisor_metrics[0]
            if cadvisor_metrics['has_memory'] and cadvisor_metrics['memory']:
                self._parse_memory(cadvisor_metrics['memory'],
                                   container_dimensions,
                                   pod_key,
                                   pod_metrics,
                                   memory_limit_map)
            if cadvisor_metrics['has_filesystem'] and 'filesystem' in cadvisor_metrics \
                    and cadvisor_metrics['filesystem']:
                self._parse_filesystem(cadvisor_metrics['filesystem'],
                                       container_dimensions)
            if cadvisor_metrics['has_network'] and cadvisor_metrics['network']:
                self._parse_network(cadvisor_metrics['network'],
                                    container_dimensions,
                                    pod_key,
                                    pod_metrics)
            if cadvisor_metrics['has_cpu'] and cadvisor_metrics['cpu']:
                self._parse_cpu(cadvisor_metrics['cpu'],
                                container_dimensions,
                                pod_key,
                                pod_metrics)
            if cadvisor_metrics['has_diskio'] and cadvisor_metrics['diskio']:
                self._parse_diskio(cadvisor_metrics['diskio'],
                                   container_dimensions,
                                   pod_key,
                                   pod_metrics)
        self.send_pod_metrics(pod_metrics, pod_dimension_map)

    def send_pod_metrics(self, pod_metrics_map, pod_dimension_map):
        for pod_key, pod_metrics in pod_metrics_map.items():
            pod_dimensions = pod_dimension_map[pod_key]
            for metric_name, metric_value in pod_metrics.items():
                self._send_metrics("pod." + metric_name,
                                   metric_value,
                                   pod_dimensions,
                                   METRIC_TYPES_UNITS[metric_name][0],
                                   METRIC_TYPES_UNITS[metric_name][1])



# (C) Copyright 2017 Hewlett Packard Enterprise Development LP
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import requests

from urllib.parse import urlparse
from urllib.parse import urlunparse

from monasca_agent.collector.checks import AgentCheck
from monasca_agent.collector.checks import utils
from monasca_agent.common.util import rollup_dictionaries

DEFAULT_TIMEOUT = "3"

# format for METRICS: (cadvisor metric name, [metric types], [metric units])
METRICS = {
    "cpu_metrics": {
        "system": ("system_time", ["gauge", "rate"],
                   ["core_seconds", "core_seconds_per_second"]),
        "total": ("total_time", ["gauge", "rate"],
                  ["core_seconds", "core_seconds_per_second"]),
        "user": ("user_time", ["gauge", "rate"],
                 ["core_seconds", "core_seconds_per_second"])
    },
    "memory_metrics": {
        "swap": ("swap_bytes", ["gauge"], ["bytes"]),
        "cache": ("cache_bytes", ["gauge"], ["bytes"]),
        "usage": ("used_bytes", ["gauge"], ["bytes"]),
        "working_set": ("working_set", ["gauge"], ["bytes"])
    },
    "filesystem_metrics": {
        "capacity": ("total_bytes", ["gauge"], ["bytes"]),
        "usage": ("usage_bytes", ["gauge"], ["bytes"])
    },
    'network_metrics': {
        "rx_bytes": ("in_bytes", ["gauge", "rate"],
                     ["total_bytes", "total_bytes_per_second"]),
        "tx_bytes": ("out_bytes", ["gauge", "rate"],
                     ["total_bytes", "total_bytes_per_second"]),
        "rx_packets": ("in_packets", ["gauge", "rate"],
                       ["total_packets", "total_packets_per_second"]),
        "tx_packets": ("out_packets", ["gauge", "rate"],
                       ["total_packets", "total_packets_per_second"]),
        "rx_dropped": ("in_dropped_packets", ["gauge", "rate"],
                       ["total_packets", "total_packets_per_second"]),
        "tx_dropped": ("out_dropped_packets", ["gauge", "rate"],
                       ["total_packets", "total_packets_per_second"]),
        "rx_errors": ("in_errors", ["gauge", "rate"],
                      ["total_errors", "total_errors_per_second"]),
        "tx_errors": ("out_errors", ["gauge", "rate"],
                      ["total_errors", "total_errors_per_second"])
    },
    "diskio_metrics": {
        "Read": ("read_bytes", ["gauge", "rate"],
                 ["total_bytes", "total_bytes_per_second"]),
        'Write': ("write_bytes", ["gauge", "rate"],
                  ["total_bytes", "total_bytes_per_second"])
    },
}


class CadvisorHost(AgentCheck):
    """Queries given cAdvisor API for node metrics
    """

    def __init__(self, name, init_config, agent_config, instances=None):
        AgentCheck.__init__(self, name, init_config, agent_config, instances)
        if instances is not None and len(instances) > 1:
            raise Exception('cAdvisor host check only supports one configured'
                            ' instance.')
        self.connection_timeout = int(init_config.get('connection_timeout',
                                                      DEFAULT_TIMEOUT))
        self.cadvisor_url = None
        self.cadvisor_machine_url = None
        self.total_mem = 0
        self.num_cores = 0

    def _parse_machine_info(self, machine_info):
        topo_info = machine_info['topology']
        # Grab first set of info from return data
        topo_info = topo_info[0]
        # Store info about total machine memory
        if topo_info['memory']:
            self.total_mem = topo_info['memory']
            self.log.debug("host memory = {}".format(self.total_mem))
        else:
            self.log.warn("Failed to retrieve host memory size")
        # Store information about number of cores (incl. threads)
        if machine_info['num_cores']:
            self.num_cores = int(machine_info['num_cores'])
            self.log.debug("number of cores of machine: {}".format(self.num_cores))
        else:
            self.log.warn("Failed to retrieve number of cores of host")

    def check(self, instance):
        if not self.cadvisor_url:
            cadvisor_url = instance.get("cadvisor_url", None)
            detect_cadvisor_url = instance.get("kubernetes_detect_cadvisor", False)
            if not cadvisor_url:
                if detect_cadvisor_url:
                    kubernetes_connector = utils.KubernetesConnector(self.connection_timeout)
                    host = kubernetes_connector.get_agent_pod_host()
                    cadvisor_url = "http://{}:4194".format(host)
                else:
                    exception_message = "Either cAdvisor url or kubernetes " \
                                        "detect cAdvisor must be set when " \
                                        "monitoring a Kubernetes Node."
                    self.log.error(exception_message)
                    raise Exception(exception_message)
            self.cadvisor_url = "{}/{}".format(cadvisor_url, "api/v2.0/stats?count=1")
        dimensions = self._set_dimensions(None, instance)
        try:
            host_metrics = requests.get(self.cadvisor_url, self.connection_timeout).json()
        except Exception as e:
            self.log.error("Error communicating with cAdvisor to collect data - {}".format(e))
        else:
            # Retrieve machine info only once
            if not self.cadvisor_machine_url:
                # Replace path in current cadvisor_url
                result = urlparse(self.cadvisor_url)
                self.cadvisor_machine_url = urlunparse(result._replace(path="api/v2.0/machine"))
                try:
                    machine_info = requests.get(self.cadvisor_machine_url).json()
                except Exception as ex:
                    self.log.error(
                        "Error communicating with cAdvisor to collect machine data - {}".format(ex))
                else:
                    self._parse_machine_info(machine_info)
            self._parse_send_metrics(host_metrics, dimensions)

    def _send_metrics(self, metric_name, value, dimensions, metric_types,
                      metric_units):
        for metric_type in metric_types:
            if metric_type == 'rate':
                dimensions.update({'unit': metric_units[metric_types.index('rate')]})
                self.rate(metric_name + "_sec", value, dimensions)
            elif metric_type == 'gauge':
                dimensions.update({'unit': metric_units[metric_types.index('gauge')]})
                self.gauge(metric_name, value, dimensions)

    def _parse_memory(self, memory_data, dimensions):
        memory_metrics = METRICS['memory_metrics']
        used_mem = -1
        for cadvisor_key, (metric_name, metric_types, metric_units) in memory_metrics.items():
            if cadvisor_key in memory_data:
                self._send_metrics("mem." + metric_name,
                                   memory_data[cadvisor_key],
                                   dimensions,
                                   metric_types, metric_units)
                if cadvisor_key == "usage":
                    used_mem = int(memory_data[cadvisor_key])
        # Calculate memory used percent
        if used_mem < 0:
            self.log.warn("no value for used memory, memory usage (percent) couldn't be calculated")
        elif self.total_mem > 0:
            used_mem_perc = (float(used_mem) / float(self.total_mem)) * 100
            # Send metric percent used
            self._send_metrics("mem.used_perc",
                               used_mem_perc,
                               dimensions,
                               metric_types, metric_units)

    def _parse_filesystem(self, filesystem_data, dimensions):
        filesystem_metrics = METRICS['filesystem_metrics']
        for filesystem in filesystem_data:
            file_dimensions = dimensions.copy()
            file_dimensions['device'] = filesystem['device']
            usage_fs = -1
            capacity_fs = 0
            for cadvisor_key, (metric_name, metric_types,
                               metric_units) in filesystem_metrics.items():
                if cadvisor_key in filesystem:
                    self._send_metrics("fs." + metric_name,
                                       filesystem[cadvisor_key],
                                       file_dimensions,
                                       metric_types, metric_units)
                if cadvisor_key == "usage":
                    usage_fs = int(filesystem[cadvisor_key])
                elif cadvisor_key == "capacity":
                    capacity_fs = int(filesystem[cadvisor_key])
            if usage_fs < 0:
                self.log.warn(
                    "no value for usage size of {}, file system usage (percent) "
                    "couldn't be calculated".format(
                        filesystem['device']))
            elif capacity_fs > 0:
                self._send_metrics("fs.usage_perc",
                                   (float(usage_fs) / capacity_fs) * 100,
                                   file_dimensions,
                                   ["gauge"], ["percent"])
            else:
                self.log.warn(
                    "no value for capacity of {}, file system usage (percent) "
                    "couldn't be calculated".format(
                        filesystem['device']))

    def _parse_network(self, network_data, dimensions):
        network_interfaces = network_data['interfaces']
        network_metrics = METRICS['network_metrics']
        interface_sum = {}
        # This function is to roll up network metrics for different interfaces
        for interface in network_interfaces:
            interface_sum = rollup_dictionaries(interface_sum, interface)

        network_dimensions = dimensions.copy()
        for cadvisor_key, (metric_name, metric_types, metric_units) in network_metrics.items():
            if cadvisor_key in interface_sum:
                self._send_metrics("net." + metric_name,
                                   interface_sum[cadvisor_key],
                                   network_dimensions,
                                   metric_types,
                                   metric_units)

    def _parse_diskio(self, diskio_data, dimensions):
        diskio_metrics = METRICS['diskio_metrics']
        disk_io_sum = {}
        for io_data in diskio_data['io_service_bytes']:
            disk_io_sum = rollup_dictionaries(disk_io_sum, io_data['stats'])

        for cadvisor_key, (metric_name, metric_types, metric_units) in diskio_metrics.items():
            if cadvisor_key in disk_io_sum:
                self._send_metrics("io." + metric_name, disk_io_sum[cadvisor_key],
                                   dimensions,
                                   metric_types,
                                   metric_units)

    def _parse_cpu(self, cpu_data, dimensions):
        cpu_metrics = METRICS['cpu_metrics']
        cpu_usage = cpu_data['usage']
        for cadvisor_key, (metric_name, metric_types, metric_units) in cpu_metrics.items():
            if cadvisor_key in cpu_usage:
                # Convert nanoseconds to seconds
                cpu_usage_sec = cpu_usage[cadvisor_key] / 1000000000.0
                self._send_metrics(
                    "cpu." + metric_name,
                    cpu_usage_sec,
                    dimensions,
                    metric_types,
                    metric_units)
                # Provide metrics for number of cores if given
                if self.num_cores > 0:
                    self._send_metrics(
                        "cpu.num_cores",
                        self.num_cores,
                        dimensions,
                        ["gauge"],
                        ["number_of_cores"])

    def _parse_send_metrics(self, metrics, dimensions):
        for host, cadvisor_metrics in metrics.items():
            host_dimensions = dimensions.copy()
            # Grab first set of metrics from return data
            cadvisor_metrics = cadvisor_metrics[0]
            if cadvisor_metrics['has_memory'] and cadvisor_metrics['memory']:
                self._parse_memory(cadvisor_metrics['memory'], host_dimensions)
            if cadvisor_metrics['has_filesystem'] and cadvisor_metrics['filesystem']:
                self._parse_filesystem(cadvisor_metrics['filesystem'], host_dimensions)
            if cadvisor_metrics['has_network'] and cadvisor_metrics['network']:
                self._parse_network(cadvisor_metrics['network'], host_dimensions)
            if cadvisor_metrics['has_diskio'] and cadvisor_metrics['diskio']:
                self._parse_diskio(cadvisor_metrics['diskio'], host_dimensions)
            if cadvisor_metrics['has_cpu'] and cadvisor_metrics['cpu']:
                self._parse_cpu(cadvisor_metrics['cpu'], host_dimensions)



# Copyright 2019 SUSE LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from datetime import datetime

from cryptography.hazmat.backends import default_backend
from cryptography import x509

from monasca_agent.collector.checks import AgentCheck


class CertificateFileCheck(AgentCheck):
    """Check the given certificate file and output a metric
       which is the number of days until it expires
    """

    def __init__(self, name, init_config, agent_config, instances=None):
        super(CertificateFileCheck, self).__init__(name, init_config,
                                                   agent_config, instances)

    def check(self, instance):
        cert_file = instance.get('cert_file', None)

        if cert_file is None:
            self.log.warning('Instance have no "cert_file" configured.')
            return

        dimensions = self._set_dimensions(None, instance)
        dimensions['cert_file'] = cert_file
        self.log.info('cert_file = %s' % cert_file)
        expire_in_days = self.get_expire_in_days(cert_file)
        if expire_in_days is not None:
            self.gauge('cert_file.cert_expire_days', expire_in_days,
                       dimensions=dimensions)
            self.log.debug('%d days till expiration for %s' % (expire_in_days,
                                                               cert_file))

    def get_expire_in_days(self, cert_file):
        """Take the path the the TLS certificate file and returns the number
           of till the certificate expires. If the certificate has already
           expired, it will return a negative number. For example,
           if the certificate has already expired 5 days prior to the check,
           -5 will be returned.
        """
        try:
            with open(cert_file, 'r') as cf:
                pem_data = cf.read().encode('ascii')

            cert = x509.load_pem_x509_certificate(pem_data, default_backend())
            return (cert.not_valid_after - datetime.utcnow()).days
        except IOError:
            self.log.warning(
                'Unable to read certificate from %s' % (cert_file))
        except ValueError:
            self.log.warning(
                'Unable to load certificate from %s. Invalid content.' % (
                    cert_file))



# (C) Copyright 2017 Hewlett Packard Enterprise Development Company LP
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import unittest

import os

import monasca_agent.common.config as configuration
import monasca_agent.common.metrics as metrics_pkg

from monasca_agent.collector.checks import AgentCheck

base_config = configuration.Config(os.path.join(os.path.dirname(__file__),
                                                'test-agent.yaml'))


class TestAgentCheck(unittest.TestCase):
    def testBadMetricKeepBatch(self):
        agent_config = base_config.get_config(sections='Main')

        check = AgentCheck("foo", {}, agent_config)

        dimensions = {'A': 'B', 'B': 'C', 'D': 'E'}
        check.submit_metric("Foo",
                            5,
                            metrics_pkg.Gauge,
                            dimensions=dimensions,
                            delegated_tenant=None,
                            hostname=None,
                            device_name=None,
                            value_meta=None)
        self.assertEqual(len(check.aggregator.metrics), 1)

        dimensions = {'A': '{}', 'B': 'C', 'D': 'E'}
        check.submit_metric("Bar",
                            5,
                            metrics_pkg.Gauge,
                            dimensions=dimensions,
                            delegated_tenant=None,
                            hostname=None,
                            device_name=None,
                            value_meta=None)
        self.assertEqual(len(check.aggregator.metrics), 1)

        dimensions = {'A': 'B', 'B': 'C', 'D': 'E'}
        check.submit_metric("Baz",
                            5,
                            metrics_pkg.Gauge,
                            dimensions=dimensions,
                            delegated_tenant=None,
                            hostname=None,
                            device_name=None,
                            value_meta=None)
        self.assertEqual(len(check.aggregator.metrics), 2)



# (C) Copyright 2017 Hewlett Packard Enterprise Development LP
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import requests

from monasca_agent.collector import checks
from monasca_agent.collector.checks import utils

DEFAULT_TIMEOUT = 5
NODE_CONDITIONS_MAP = {
    "OutOfDisk": {
        "metric_name": "node.out_of_disk",
        "expected_status": "False"
    },
    "MemoryPressure": {
        "metric_name": "node.memory_pressure",
        "expected_status": "False"
    },
    "DiskPressure": {
        "metric_name": "node.disk_pressure",
        "expected_status": "False"
    },
    "Ready": {
        "metric_name": "node.ready_status",
        "expected_status": "True"
    }
}


class KubernetesAPI(checks.AgentCheck):
    """Queries Kubernetes API to get metrics about the Kubernetes deployment
    """
    def __init__(self, name, init_config, agent_config, instances=None):
        checks.AgentCheck.__init__(self, name, init_config, agent_config, instances)
        if instances is not None and len(instances) > 1:
            raise Exception('Kubernetes api check only supports one configured instance.')
        self.connection_timeout = int(init_config.get('connection_timeout', DEFAULT_TIMEOUT))
        self.kubernetes_connector = None
        self.kubernetes_api = None

    def prepare_run(self):
        """Set up Kubernetes connection information"""
        instance = self.instances[0]
        host = instance.get("host", None)
        derive_api_url = instance.get("derive_api_url", None)
        if not host:
            if derive_api_url:
                self.kubernetes_connector = utils.KubernetesConnector(self.connection_timeout)
            else:
                exception_message = "Either Kubernetes API url (host and port)" \
                                    " or derive_api_url=True must be set" \
                                    " when running Kubernetes API plugin."
                self.log.error(exception_message)
                raise Exception(exception_message)
        else:
            kubernetes_api_port = instance.get("kubernetes_api_port", "8080")
            self.kubernetes_api = "http://{}:{}".format(host, kubernetes_api_port)

    def check(self, instance):
        kubernetes_labels = instance.get('kubernetes_labels', ["app"])
        dimensions = self._set_dimensions(None, instance)
        # Remove hostname from dimensions as the majority of the metrics are not
        # tied to the hostname.
        del dimensions['hostname']
        kubernetes_api_health = self._get_api_health()
        self.gauge("kubernetes.api.health_status", 0 if kubernetes_api_health else 1, dimensions,
                   hostname="SUPPRESS")
        self._report_cluster_component_statuses(dimensions)
        self._report_nodes_metrics(dimensions)
        self._report_deployment_metrics(dimensions, kubernetes_labels)
        self._report_replication_controller_metrics(dimensions, kubernetes_labels)

    def _send_request(self, endpoint, as_json=True):
        if self.kubernetes_connector:
            return self.kubernetes_connector.get_request(endpoint, as_json=as_json)
        else:
            result = requests.get("{}/{}".format(self.kubernetes_api, endpoint))
            return result.json() if as_json else result

    def _get_api_health(self):
        try:
            result = self._send_request("healthz", as_json=False)
        except Exception as e:
            self.log.error("Error connecting to the health endpoint with exception {}".format(e))
            return False
        else:
            # Return true if 'ok' is in result
            return 'ok' in result.iter_lines()

    def _report_cluster_component_statuses(self, dimensions):
        try:
            component_statuses = self._send_request("/api/v1/componentstatuses")
        except Exception as e:
            self.log.error("Error getting data from Kubernetes API - {}".format(e))
            return
        for component in component_statuses['items']:
            component_dimensions = dimensions.copy()
            component_dimensions['component_name'] = component['metadata']['name']
            component_status = False
            component_conditions = component['conditions']
            for condition in component_conditions:
                if 'type' in condition and condition['type'] == 'Healthy':
                    if condition['status']:
                        component_status = True
                        break
            self.gauge(
                "kubernetes.component_status",
                0 if component_status else 1,
                component_dimensions,
                hostname="SUPPRESS")

    def _set_kubernetes_dimensions(self, dimensions, type, metadata, kubernetes_labels):
        dimensions['type'] = metadata['name']
        dimensions['namespace'] = metadata['namespace']
        if 'labels' in metadata:
            labels = metadata['labels']
            for label in kubernetes_labels:
                if label in labels:
                    dimensions[label] = labels[label]

    def _report_node_resource_metrics(self, resource, metrics, node_dimensions):
        resource_metrics_dimensions = node_dimensions.copy()
        for metric_name, metric_value in metrics.items():
            if "gpu" in metric_name:
                continue
            if metric_name == "memory":
                metric_name += "_bytes"
                metric_value = utils.convert_memory_string_to_bytes(metric_value)
                resource_metrics_dimensions.update({'unit': 'bytes'})
            elif metric_name == "cpu":
                resource_metrics_dimensions.update({'unit': 'cores'})
            metric_name = "kubernetes.node.{}.{}".format(resource, metric_name)
            self.gauge(metric_name, float(metric_value), resource_metrics_dimensions)

    def _report_node_conditions_metrics(self, node_conditions, node_dimensions):
        for condition in node_conditions:
            condition_type = condition["type"]
            if condition_type in NODE_CONDITIONS_MAP:
                condition_map = NODE_CONDITIONS_MAP[condition_type]
                condition_status = condition['status']
                if condition_status == condition_map['expected_status']:
                    self.gauge("kubernetes." + condition_map['metric_name'], 0, node_dimensions)
                else:
                    value_meta = {"reason": condition['message'][:1024]}
                    self.gauge(
                        "kubernetes." +
                        condition_map['metric_name'],
                        1,
                        node_dimensions,
                        value_meta=value_meta)

    def _report_nodes_metrics(self, dimensions):
        try:
            nodes = self._send_request("/api/v1/nodes")
        except Exception as e:
            self.log.error("Error getting node data from Kubernetes API - {}".format(e))
            return
        for node in nodes['items']:
            node_dimensions = dimensions.copy()
            node_dimensions['hostname'] = node['metadata']['name']
            node_status = node['status']
            self._report_node_conditions_metrics(node_status['conditions'], node_dimensions)
            if 'spec' in node and 'unschedulable' in node['spec']:
                if node['spec']['unschedulable']:
                    continue
            node_capacity = node_status['capacity']
            node_allocatable = node_status['allocatable']
            self._report_node_resource_metrics('capacity', node_capacity, node_dimensions)
            self._report_node_resource_metrics('allocatable', node_allocatable, node_dimensions)

    def _report_deployment_metrics(self, dimensions, kubernetes_labels):
        try:
            deployments = self._send_request("/apis/extensions/v1beta1/deployments")
        except Exception as e:
            self.log.error("Error getting deployment data from Kubernetes API - {}".format(e))
            return
        for deployment in deployments['items']:
            try:
                deployment_dimensions = dimensions.copy()
                self._set_kubernetes_dimensions(
                    deployment_dimensions,
                    "deployment",
                    deployment['metadata'],
                    kubernetes_labels)
                deployment_status = deployment['status']
                deployment_replicas = deployment_status['replicas']
                deployment_updated_replicas = deployment_status['updatedReplicas']
                deployment_available_replicas = deployment_status['availableReplicas']
                deployment_unavailable_replicas = \
                    deployment_available_replicas - deployment_replicas
                self.gauge("kubernetes.deployment.replicas", deployment_replicas,
                           deployment_dimensions, hostname="SUPPRESS")
                self.gauge(
                    "kubernetes.deployment.available_replicas",
                    deployment_available_replicas,
                    deployment_dimensions,
                    hostname="SUPPRESS")
                self.gauge(
                    "kubernetes.deployment.unavailable_replicas",
                    deployment_unavailable_replicas,
                    deployment_dimensions,
                    hostname="SUPPRESS")
                self.gauge("kubernetes.deployment.updated_replicas", deployment_updated_replicas,
                           deployment_dimensions, hostname="SUPPRESS")
            except Exception as e:
                self.log.info(
                    "Error {} parsing deployment {}. Skipping".format(
                        e, deployment), exc_info=e)

    def _report_replication_controller_metrics(self, dimensions, kubernetes_labels):
        # Get namespaces first
        try:
            namespaces = self._send_request("/api/v1/namespaces")
        except Exception as e:
            self.log.error("Error getting namespaces from API - {}. "
                           "Skipping getting replication controller metrics".format(e))
            return
        for namespace in namespaces['items']:
            namespace_name = namespace['metadata']['name']
            try:
                replication_controllers = self._send_request(
                    "/api/v1/namespaces/{}/replicationcontrollers".format(namespace_name))
            except Exception as e:
                self.log.error("Error getting replication controllers for the namespace {} "
                               "with the error {}".format(namespace, e))
                continue
            if 'items' not in replication_controllers:
                continue
            for rc in replication_controllers['items']:
                rc_dimensions = dimensions.copy()
                self._set_kubernetes_dimensions(
                    rc_dimensions,
                    "replication_controller",
                    rc['metadata'],
                    kubernetes_labels)
                rc_status = rc['status']
                if 'replicas' not in rc_status or not rc_status['replicas']:
                    continue
                self.gauge("kubernetes.replication.controller.replicas", rc_status['replicas'],
                           rc_dimensions, hostname="SUPPRESS")
                self.gauge(
                    "kubernetes.replication.controller.ready_replicas",
                    rc_status['readyReplicas'],
                    rc_dimensions,
                    hostname="SUPPRESS")



# (C) Copyright 2016-2017 Hewlett Packard Enterprise Development LP
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""VCenter Only.

Generic VCenter check. This check allows you to specify particular metrics that
you want from vCenter in your configuration.
"""

from monasca_agent.collector.checks import AgentCheck
from oslo_vmware import api
from oslo_vmware import vim_util
import traceback

CLUSTER_COMPUTE_PROPERTIES = ["name", "host", "datastore"]
CLUSTER_HOST_PROPERTIES = ['runtime.connectionState',
                           'runtime.inMaintenanceMode',
                           'summary.hardware.numCpuThreads',
                           'summary.hardware.memorySize',
                           'summary.hardware.cpuMhz']
CLUSTER_PERF_COUNTERS = [
    "cpu.usagemhz.average",
    "cpu.totalmhz.average",
    "mem.usage.average",
    "mem.consumed.average",
    "mem.totalmb.average"]
STORAGE_VOLUME_PROPERTIES = ["host",
                             "summary.capacity",
                             "summary.freeSpace",
                             "summary.name",
                             "summary.multipleHostAccess",
                             "summary.accessible",
                             "summary.maintenanceMode",
                             "summary.type"]

CPU_TOTAL_MHZ = "vcenter.cpu.total_mhz"
CPU_USED_MHZ = "vcenter.cpu.used_mhz"
CPU_USED_PERCENT = "vcenter.cpu.used_perc"
CPU_TOTAL_LOGICAL_CORES = "vcenter.cpu.total_logical_cores"
MEMORY_TOTAL_MB = "vcenter.mem.total_mb"
MEMORY_USED_MB = "vcenter.mem.used_mb"
MEMORY_USED_PERCENT = "vcenter.mem.used_perc"
DISK_TOTAL_SPACE_MB = "vcenter.disk.total_space_mb"
DISK_TOTAL_USED_SPACE_MB = "vcenter.disk.total_used_space_mb"
DISK_TOTAL_USED_SPACE_PERCENT = "vcenter.disk.total_used_space_perc"

PERF_MANAGER_TYPE = "PerformanceManager"
PERF_RESOURCE = 'ClusterResource'
NUM_SAMPLES = 15
SAMPLE_RATE = 300
ALLOWED_DATASTORE_TYPES = ['VMFS', 'NFS']


class VCenterCheck(AgentCheck):

    def __init__(self, name, init_config, agent_config):
        AgentCheck.__init__(self, name, init_config, agent_config)
        self._max_objects = 1000
        self.session = None
        self.is_new_session = True
        self._resource_moid_dict = {}

    def stop(self):
        """To be executed when the agent is being stopped to clean resources.
        """
        if self.session is not None:
            self.session.logout()
            self.session = None
            self.is_new_session = True

    def _propset_dict(self, propset):
        """Turns a propset list into a dictionary

        PropSet is an optional attribute on ObjectContent objects
        that are returned by the VMware API.
        :param propset: a property 'set' from ObjectContent
        :return: dictionary representing property set
        """
        if propset is None:
            return {}
        return {prop.name: prop.val for prop in propset}

    def _get_api_session(self):
        api_session = api.VMwareAPISession(
            self.vcenter_ip,
            self.user,
            self.password,
            3,  # retry count
            0.5,  # task_poll_interval
            port=self.port,
            scheme="https")
        return api_session

    def _find_entity_mor(self, entity_list, entity_name):
        for object_contents in entity_list:
            for obj_content in object_contents[1]:
                for dyn_props in obj_content.propSet:
                    if dyn_props.val == entity_name.decode('utf-8'):
                        return obj_content

    def _get_resource_dict_by_name(self, resource_name):
        """Return moid from the specified resource"""
        self._get_cluster_info_dict()
        self.log.debug("Loaded MORs for clusters %s" %
                       str(self._resource_moid_dict.keys()))
        return self._resource_moid_dict.get(resource_name)

    def _build_resource_dict(self, object_content):
        """Method to build the resource dictionary. Cluster_name is the key,
        and its host and datastore names are the values
        {'cluster_name': {
            'moid': 'cluster1',
            'host': ['h1', 'h2'],
            'datastore': ['ds1', 'ds2']
            }
        }
        """
        cluster_moid = str(object_content.obj.value)
        # extract host and datastore names
        host_names = []
        datastore_names = []
        dict_key = None
        propSet = object_content.propSet
        for dynamic_prop in propSet:
            mor_array = dynamic_prop.val
            if dynamic_prop.name == 'name':
                dict_key = dynamic_prop.val
            else:
                for obj in mor_array:
                    for mor in obj[1]:
                        if dynamic_prop.name == 'datastore':
                            datastore_names.append(str(mor.value))
                        elif dynamic_prop.name == 'host':
                            host_names.append(str(mor.value))
        if dict_key:
            self._resource_moid_dict[str(dict_key)] = {
                'moid': cluster_moid,
                'host': host_names,
                'datastore': datastore_names,
            }

    def _get_cluster_info_dict(self):
        """Load configured cluster moids from the vCenter"""
        self.log.debug("Inside _get_cluster_info_dict")
        if self.session is None:
            self.session = self._get_api_session()
        result = self.session.invoke_api(vim_util,
                                         "get_objects",
                                         self.session.vim,
                                         "ClusterComputeResource",
                                         self._max_objects,
                                         CLUSTER_COMPUTE_PROPERTIES)
        for name in self.clusters:
            mor = self._find_entity_mor(result, name)
            if mor:
                self._build_resource_dict(mor)

    def _get_sample(self, samples, counter_name, is_summation=False):
        res = 0
        num_samples = 0

        for cn in samples:
            if cn.startswith(counter_name):
                vals = samples[cn]
                if vals:
                    for val in vals:
                        i_val = int(val)
                        if i_val != -1:
                            res += i_val
                            num_samples += 1

        if not is_summation and num_samples:
            res /= num_samples

        return res

    def _get_shared_datastores(self, datastore_stats, managed_cluster):
        """Method to find the shared datastores associated with the cluster
        Also handles cluster having single host with a non-shared datastore.
        """
        shared_datastores_ids = []
        res_dict = self._get_resource_dict_by_name(managed_cluster)
        host_set = set(res_dict.get('host'))
        for object_contents in datastore_stats:
            for object_content in object_contents[1]:
                ds_mor = object_content.obj.value
                datastore_host_ids = []
                if ds_mor in res_dict.get('datastore'):
                    propSets = object_content.propSet
                    for propSet in propSets:
                        if propSet.name == 'host':
                            ds_hm_array = propSet.val
                            host_mounts = ds_hm_array[0]
                            for host_mount in host_mounts:
                                datastore_host_ids.append(host_mount.key.value)
                    if host_set.issubset(set(datastore_host_ids)):
                        shared_datastores_ids.append(ds_mor)
                    self.log.debug("Cluster host list ==" + str(host_set))
                    self.log.debug("Datastore host list ==" +
                                   str(datastore_host_ids))
        self.log.debug("shared_datastores_ids==" +
                       str(shared_datastores_ids))
        return shared_datastores_ids

    def _is_valid_datastore(self, propSets):
        propdict = self._propset_dict(propSets)
        return (propdict.get('summary.accessible') and
                (propdict.get('summary.maintenanceMode') is None or
                 propdict.get('summary.maintenanceMode') == 'normal') and
                propdict['summary.type'] in ALLOWED_DATASTORE_TYPES)

    def _process_storage_data(self, datastore_stats, managed_cluster):
        shared_ds = self._get_shared_datastores(datastore_stats,
                                                managed_cluster)
        capacity = 0
        freeSpace = 0
        self.log.info("Polling for the datastores: " + str(shared_ds))
        for object_contents in datastore_stats:
            for object_content in object_contents[1]:
                ds_mor = object_content.obj.value
                if ds_mor in shared_ds:
                    propSets = object_content.propSet
                    if self._is_valid_datastore(propSets):
                        for propSet in propSets:
                            if propSet.name == 'summary.capacity':
                                self.log.debug("Calculating capacity "
                                               "of datastore: %s in cluster: "
                                               "%s" %
                                               (ds_mor, managed_cluster))
                                capacity += int(propSet.val)
                            elif propSet.name == 'summary.freeSpace':
                                self.log.debug("Calculating freeSpace of "
                                               "datastore: %s in cluster: %s"
                                               % (ds_mor, managed_cluster))
                                freeSpace += int(propSet.val)
        usedSpace = capacity - freeSpace
        self.log.debug("Total capacity:" + str(capacity) +
                       " used:" + str(usedSpace) + " free:" + str(freeSpace))
        return (capacity, usedSpace)

    def _get_properties_for_a_collection_of_objects(self, vim, type, obj_list,
                                                    properties):
        """Gets the list of properties for the collection of
        objects of the type specified.
        """
        client_factory = vim.client.factory
        if len(obj_list) == 0:
            return []
        prop_spec = vim_util.build_property_spec(client_factory,
                                                 type,
                                                 properties)
        lst_obj_specs = []
        for obj in obj_list:
            lst_obj_specs.append(vim_util.build_object_spec(
                client_factory,
                obj, []))
        prop_filter_spec = vim_util.build_property_filter_spec(client_factory,
                                                               [prop_spec],
                                                               lst_obj_specs)
        options = client_factory.create('ns0:RetrieveOptions')
        options.maxObjects = self._max_objects
        return vim.RetrievePropertiesEx(
            vim.service_content.propertyCollector,
            specSet=[prop_filter_spec],
            options=options)

    def _is_valid_host(self, propSets):
        propdict = self._propset_dict(propSets)
        return ((propdict.get('runtime.inMaintenanceMode') is not True) and
                propdict.get('runtime.connectionState') == 'connected')

    def _process_host_data(self, host_stats, managed_cluster):
        shared_hosts = self._find_entity_mor(host_stats, managed_cluster)
        cluster_hosts_data = {'numCpuThreads': 0,
                              'memorySizeMb': 0,
                              'cpuMhz': 0}
        self.log.debug("Polling for the hosts: %s" % (str(shared_hosts)))
        if shared_hosts:
            propSets = shared_hosts.propSet
            for propSet in propSets:
                if propSet.name == 'host':
                    host_ret = propSet.val
                    host_mors = host_ret.ManagedObjectReference
        if host_mors:
            method_call = '_get_properties_for_a_collection_of_objects'
            result = self.session.invoke_api(self,
                                             method_call,
                                             self.session.vim,
                                             "HostSystem",
                                             host_mors,
                                             CLUSTER_HOST_PROPERTIES)
        for obj_contents in result:
            for obj_content in obj_contents[1]:
                propSets = obj_content.propSet
                if self._is_valid_host(propSets):
                    for propSet in propSets:
                        if propSet.name == "summary.hardware.numCpuThreads":
                            cluster_hosts_data['numCpuThreads'] += propSet.val
                        elif propSet.name == 'summary.hardware.memorySize':
                            cluster_hosts_data['memorySizeMb'] += int(
                                (propSet.val) / (1024 * 1024))
                        elif propSet.name == 'summary.hardware.cpuMhz':
                            cluster_hosts_data['cpuMhz'] += propSet.val
        return cluster_hosts_data

    def check(self, instance):
        try:
            if self.is_new_session:
                self.instance = instance
                self.vcenter_ip = instance.get('vcenter_ip', None)
                self.user = instance.get('username', None)
                self.password = instance.get('password', None)
                self.port = instance.get('port', 443)
                self.clusters = instance.get('clusters', None)
                if not self.vcenter_ip:
                    self.log.warn("vCenter not configured")
                    return
                if not self.clusters:
                    self.log.warn("No clusters configured to monitor")
                    return
                self.session = self._get_api_session()
                self.vc_uuid = (self.session.vim.service_content.
                                about.instanceUuid)
                self._ops = VcenterOperations(self.session,
                                              self._max_objects,
                                              self.log)
                self._ops._properties_updated_event()
                self.perf_query_specs = self._ops._get_perf_query_spec(
                    CLUSTER_PERF_COUNTERS,
                    SAMPLE_RATE,
                    NUM_SAMPLES)
                self.is_new_session = False
            for managed_cluster in self.clusters:
                try:
                    res_dict = self._get_resource_dict_by_name(managed_cluster)
                    if not res_dict:
                        self.log.error("Configured cluster not found "
                                       "in vCenter")
                        continue
                    self.log.debug("=" * 30 + " Start of cycle" + "=" * 30)
                    self.log.debug("Collecting for cluster: %s (%s)" %
                                   (managed_cluster,
                                    str(res_dict.get('moid'))))
                    cluster_samples = self._ops.\
                        _get_perf_stats(res_dict,
                                        self.perf_query_specs)
                    if not cluster_samples:
                        self.log.error("Activated cluster not found in"
                                       " vCenter")
                        continue

                    host_stats = self._ops._get_host_stats_from_cluster()
                    cluster_host_stats = self._process_host_data(
                        host_stats,
                        managed_cluster)
                    cpu_total_logical_cores = cluster_host_stats.get(
                        'numCpuThreads')

                    cpu_usage = self._get_sample(cluster_samples,
                                                 'cpu.usagemhz.average')
                    cpu_total = self._get_sample(cluster_samples,
                                                 'cpu.totalmhz.average')
                    cpu_usage_percent = float(0.0)
                    if cpu_total > 0:
                        cpu_usage_percent = round(
                            (float(cpu_usage) / cpu_total) * 100, 2)

                    # Considering host level stats for RAM
                    mem_mb_total = cluster_host_stats.get('memorySizeMb')
                    mem_kb_consumed = self._get_sample(cluster_samples,
                                                       'mem.consumed.average')
                    # convert this to MB
                    mem_mb_consumed = mem_kb_consumed / 1024
                    mem_used_percent = float(0.0)
                    if mem_mb_total > 0:
                        mem_used_percent = round(
                            (float(mem_mb_consumed) / mem_mb_total) * 100, 2)

                    datastore_stats = self._ops._read_storage_data()
                    (capacity_bytes, usedSpace_bytes) = \
                        self._process_storage_data(
                            datastore_stats,
                            managed_cluster)
                    storage_used_percent = float(0.0)
                    if capacity_bytes > 0:
                        storage_used_percent = round(
                            float(usedSpace_bytes) / capacity_bytes, 2) * 100
                    capacity_mb = capacity_bytes / (1024 * 1024)
                    usedSpace_mb = usedSpace_bytes / (1024 * 1024)

                    self.log.debug(managed_cluster +
                                   ", CPU used (avg MHz): %s" % str(cpu_usage))
                    self.log.debug(managed_cluster +
                                   ", CPU total (avg MHz): %s" %
                                   str(cpu_total))
                    self.log.debug(managed_cluster +
                                   ", Memory total (MB): %s" %
                                   str(mem_mb_total))
                    self.log.debug(managed_cluster +
                                   ", Memory consumed (MB): %s" %
                                   str(mem_mb_consumed))
                    self.log.debug(managed_cluster +
                                   ", Storage total (MB): %s" %
                                   str(capacity_mb))
                    self.log.debug(managed_cluster +
                                   ", Storage used (MB): %s" %
                                   str(usedSpace_mb))
                    self.log.debug(managed_cluster +
                                   ", CPU cores : %s" %
                                   str(cpu_total_logical_cores))
                    self.log.debug(managed_cluster +
                                   ", CPU cores: %s" %
                                   str(cpu_total_logical_cores) +
                                   ", CPU usage percent: %s" %
                                   str(cpu_usage_percent) +
                                   ", Memory usage percent: %s" %
                                   str(mem_used_percent) +
                                   ", Storage usage percent: %s" %
                                   str(storage_used_percent))

                    data = {
                        CPU_TOTAL_MHZ: cpu_total,
                        CPU_USED_MHZ: cpu_usage,
                        CPU_USED_PERCENT: cpu_usage_percent,
                        CPU_TOTAL_LOGICAL_CORES: cpu_total_logical_cores,
                        MEMORY_TOTAL_MB: mem_mb_total,
                        MEMORY_USED_MB: mem_mb_consumed,
                        MEMORY_USED_PERCENT: mem_used_percent,
                        DISK_TOTAL_SPACE_MB: capacity_mb,
                        DISK_TOTAL_USED_SPACE_MB: usedSpace_mb,
                        DISK_TOTAL_USED_SPACE_PERCENT: storage_used_percent
                    }

                    self._set_metrics(managed_cluster,
                                      data,
                                      res_dict.get("moid"))
                    self.log.debug("=" * 30 + " End of cycle" + "=" * 30)
                except Exception:
                    self.log.error(traceback.format_exc())
                    self.log.error("Exception occurred while polling for %s"
                                   % managed_cluster)
        except Exception as e:
            self.is_new_session = False
            self.log.error("Exception: %s" % str(e))
            try:
                self.session.close()
            except Exception:
                self.log.error(traceback.format_exc())
            raise e

    def _set_metrics(self, managed_cluster, data, mor_id):
        dimensions = self._get_dims(managed_cluster, mor_id)

        for key in data.keys():
            self.gauge(key, data.get(key), dimensions=dimensions)
            self.log.debug("Post metric data for %s,  %s: %d" %
                           (managed_cluster, key, data.get(key)))

    def _get_dims(self, cluster, mor_id):
        cluster_id = mor_id + "." + self.vc_uuid
        local_dimensions = {
            "vcenter_ip": self.vcenter_ip,
            "esx_cluster_id": cluster_id
        }
        final_dimensions = self._set_dimensions(local_dimensions,
                                                self.instance)
        return final_dimensions

    @staticmethod
    def parse_agent_config(agentConfig):
        if not agentConfig.get('vcenter_ip'):
            return False

        vcenter_ip = agentConfig.get('vcenter_ip')
        user = agentConfig.get('username')
        paswd = agentConfig.get('password')
        config = {
            'instances': [{
                'vcenter_ip': vcenter_ip,
                'username': user,
                'password': paswd,
            }]
        }
        return config


class VcenterOperations(object):
    """Class to invoke vCenter APIs calls.

    vCenter APIs calls are required by various pollsters, collecting data from
    VMware infrastructure.
    """
    def __init__(self, session, max_objects, logger):
        self.session = session
        self.max_objects = max_objects
        self.log = logger
        self.counters = {}

    def _get_perf_counters(self):
        try:
            self.log.info("Loading VmWare performance counters")
            perf_counters = []
            property_dict = {'perf_counter': ['perfCounter']}

            prop_collector = self.session.vim.service_content.propertyCollector
            self.client_factory = self.session.vim.client.factory
            perf_manager = self.session.vim.service_content.perfManager
            options = self.client_factory.create('ns0:RetrieveOptions')
            options.maxObjects = 1

            prop_spec = vim_util.build_property_spec(
                self.client_factory,
                PERF_MANAGER_TYPE,
                property_dict.get('perf_counter'))
            obj_spec = vim_util.build_object_spec(
                self.client_factory,
                perf_manager,
                None)

            filter_spec = vim_util.build_property_filter_spec(
                self.client_factory, [prop_spec], [obj_spec])

            object_contents = self.session.invoke_api(
                self.session.vim,
                "RetrievePropertiesEx",
                prop_collector,
                specSet=[filter_spec],
                options=options)

            if object_contents is not None:
                for object_content in object_contents:
                    dynamic_properties = object_content[1][0].propSet
                    for dynamic_property in dynamic_properties:
                        perf_counters = dynamic_property.val.PerfCounterInfo

            return perf_counters
        except Exception as ex:
            self.log.error("Exception in _get_perf_counters: %s"
                           % str(ex.message))

    def _properties_updated_event(self):
        perfCounters = self._get_perf_counters()

        if perfCounters is None:
            return

        for perfCounter in perfCounters:
            counter_id = perfCounter.key
            counter_group = perfCounter.groupInfo.key
            counter_name = perfCounter.nameInfo.key
            counter_rollup_type = str(perfCounter.rollupType)
            full_counter_name = counter_group + "." + \
                counter_name + "." + counter_rollup_type
            if (full_counter_name in CLUSTER_PERF_COUNTERS):
                self.counters[full_counter_name] = perfCounter
                self.counters[counter_id] = full_counter_name

    def _get_perf_counter_ids(self, counter_names):
        res = []
        for counter_name in counter_names:
            perf_counter_info = self.counters.get(counter_name)
            if perf_counter_info is not None:
                res.append(perf_counter_info.key)
        return res

    def _get_perf_query_spec(self,
                             counter_names,
                             sample_rate,
                             num_samples):
        counter_ids = self._get_perf_counter_ids(counter_names)

        perf_metric_ids = []
        for cid in counter_ids:
            metric = self.client_factory.create('ns0:PerfMetricId')
            metric.counterId = cid
            metric.instance = '*'
            perf_metric_ids.append(metric)
        return perf_metric_ids

    def _get_perf_stats(self, entity, perf_metric_ids):
        perf_query_spec = self.client_factory.create('ns0:PerfQuerySpec')
        perf_query_spec.entity = entity.get('moid')
        perf_query_spec.metricId = perf_metric_ids
        perf_query_spec.intervalId = SAMPLE_RATE
        perf_query_spec.maxSample = NUM_SAMPLES
        perf_manager = self.session.vim.service_content.perfManager
        perf_entity_metric_base = self.session.invoke_api(
            self.session.vim,
            'QueryPerf',
            perf_manager,
            querySpec=[perf_query_spec])

        perf_result = {}
        if perf_entity_metric_base:
            for i in range(0, len(perf_entity_metric_base)):
                perf_entity_metric_csv = perf_entity_metric_base[i]
                perf_metric_series_csvs = []
                perf_metric_series_csvs = perf_entity_metric_csv.value

                if perf_metric_series_csvs is None or \
                        len(perf_metric_series_csvs) == 0:
                    continue

                for j in range(0, len(perf_metric_series_csvs)):
                    perf_metric_series_csv = perf_metric_series_csvs[j]
                    name = self.counters[perf_metric_series_csv.id.counterId]
                    instance = perf_metric_series_csv.id.instance
                    if (instance is not None and
                        len(instance) > 0 and
                            instance != "*"):
                        name += "." + instance
                    perf_result[name] = perf_metric_series_csv.value

        return perf_result

    def _read_storage_data(self):
        datastore_stats = self.session.invoke_api(
            vim_util,
            "get_objects",
            self.session.vim,
            "Datastore",
            self.max_objects,
            STORAGE_VOLUME_PROPERTIES
        )
        return datastore_stats

    def _get_host_stats_from_cluster(self):
        prop_dict = self.session.invoke_api(vim_util,
                                            "get_objects",
                                            self.session.vim,
                                            "ClusterComputeResource",
                                            self.max_objects,
                                            ['host', 'name'])
        return prop_dict



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import unittest
from tests.common import load_check
from nose.plugins.skip import SkipTest


class GearmanTestCase(unittest.TestCase):

    def testMetrics(self):
        raise SkipTest('Requires Gearman')

        config = {
            'instances': [{
                'dimensions': {'first': 'second'}
            }]
        }
        agent_config = {
            'version': '0.1',
            'api_key': 'toto'
        }

        self.check = load_check('gearmand', config, agent_config)

        self.check.check(config['instances'][0])

        metrics = self.check.get_metrics()
        self.assertIsInstance(metrics, list)
        self.assertTrue(len(metrics) == 4)
        self.assertTrue(len([k for k in metrics if "second" in k[3]['dimensions']['first']]) == 4)
